# This file is generated from pydcs_export.lua

from . import unittype


class Fortification:

    class Command_Center(unittype.StaticType):
        id = ".Command Center"
        name = "Command Center"
        shape_name = "ComCenter"
        rate = 100

    class Hangar_A(unittype.StaticType):
        id = "Hangar A"
        name = "Hangar A"
        shape_name = "angar_a"
        rate = 100

    class Tech_hangar_A(unittype.StaticType):
        id = "Tech hangar A"
        name = "Tech hangar A"
        shape_name = "ceh_ang_a"
        rate = 100

    class Farm_A(unittype.StaticType):
        id = "Farm A"
        name = "Farm A"
        shape_name = "ferma_a"
        rate = 100

    class Farm_B(unittype.StaticType):
        id = "Farm B"
        name = "Farm B"
        shape_name = "ferma_b"
        rate = 100

    class Garage_A(unittype.StaticType):
        id = "Garage A"
        name = "Garage A"
        shape_name = "garage_a"
        rate = 100

    class Boiler_house_A(unittype.StaticType):
        id = "Boiler-house A"
        name = "Boiler-house A"
        shape_name = "kotelnaya_a"
        rate = 100

    class Restaurant_1(unittype.StaticType):
        id = "Restaurant 1"
        name = "Restaurant 1"
        shape_name = "restoran1"
        rate = 100

    class Comms_tower_M(unittype.StaticType):
        id = "Comms tower M"
        name = "Comms tower M"
        shape_name = "tele_bash_m"
        rate = 100

    class Cafe(unittype.StaticType):
        id = "Cafe"
        name = "Cafe"
        shape_name = "stolovaya"
        rate = 100

    class Workshop_A(unittype.StaticType):
        id = "Workshop A"
        name = "Workshop A"
        shape_name = "tec_a"
        rate = 100

    class Electric_power_box(unittype.StaticType):
        id = "Electric power box"
        name = "Electric power box"
        shape_name = "tr_budka"
        rate = 100

    class Supermarket_A(unittype.StaticType):
        id = "Supermarket A"
        name = "Supermarket A"
        shape_name = "uniwersam_a"
        rate = 100

    class Water_tower_A(unittype.StaticType):
        id = "Water tower A"
        name = "Water tower A"
        shape_name = "wodokachka_a"
        rate = 100

    class TV_tower(unittype.StaticType):
        id = "TV tower"
        name = "TV tower"
        shape_name = "tele_bash"
        rate = 100

    class Shelter(unittype.StaticType):
        id = "Shelter"
        name = "Shelter"
        shape_name = "ukrytie"
        rate = 100

    class Repair_workshop(unittype.StaticType):
        id = "Repair workshop"
        name = "Repair workshop"
        shape_name = "tech"
        rate = 100

    class Railway_station(unittype.StaticType):
        id = "Railway station"
        name = "Railway station"
        shape_name = "r_vok_sd"
        rate = 100

    class Railway_crossing_A(unittype.StaticType):
        id = "Railway crossing A"
        name = "Railway crossing A"
        shape_name = "pereezd_big"
        rate = 100

    class Railway_crossing_B(unittype.StaticType):
        id = "Railway crossing B"
        name = "Railway crossing B"
        shape_name = "pereezd_small"
        rate = 100

    class WC(unittype.StaticType):
        id = "WC"
        name = "WC"
        shape_name = "WC"
        rate = 100

    class Small_house_1A_area(unittype.StaticType):
        id = "Small house 1A area"
        name = "Small house 1A area"
        shape_name = "domik1a-all"
        rate = 100

    class Small_house_1A(unittype.StaticType):
        id = "Small house 1A"
        name = "Small house 1A"
        shape_name = "domik1a"
        rate = 100

    class Small_house_1B_area(unittype.StaticType):
        id = "Small house 1B area"
        name = "Small house 1B area"
        shape_name = "domik1b-all"
        rate = 100

    class Small_house_1B(unittype.StaticType):
        id = "Small house 1B"
        name = "Small house 1B"
        shape_name = "domik1b"
        rate = 100

    class Small_house_1C_area(unittype.StaticType):
        id = "Small house 1C area"
        name = "Small house 1C area"
        shape_name = "dom2c-all"
        rate = 100

    class Small_house_2C(unittype.StaticType):
        id = "Small house 2C"
        name = "Small house 2C"
        shape_name = "dom2c"
        rate = 100

    class Shop(unittype.StaticType):
        id = "Shop"
        name = "Shop"
        shape_name = "magazin"
        rate = 100

    class Tech_combine(unittype.StaticType):
        id = "Tech combine"
        name = "Tech combine"
        shape_name = "kombinat"
        rate = 100

    class Chemical_tank_A(unittype.StaticType):
        id = "Chemical tank A"
        name = "Chemical tank A"
        shape_name = "him_bak_a"
        rate = 100

    class Small_warehouse_1(unittype.StaticType):
        id = "Small werehouse 1"
        name = "Small warehouse 1"
        shape_name = "s1"
        rate = 100

    class Small_warehouse_2(unittype.StaticType):
        id = "Small werehouse 2"
        name = "Small warehouse 2"
        shape_name = "s2"
        rate = 100

    class Small_warehouse_3(unittype.StaticType):
        id = "Small werehouse 3"
        name = "Small warehouse 3"
        shape_name = "s3"
        rate = 100

    class Small_warehouse_4(unittype.StaticType):
        id = "Small werehouse 4"
        name = "Small warehouse 4"
        shape_name = "s4"
        rate = 100

    class Garage_B(unittype.StaticType):
        id = "Garage B"
        name = "Garage B"
        shape_name = "garage_b"
        rate = 100

    class Garage_small_A(unittype.StaticType):
        id = "Garage small A"
        name = "Garage small A"
        shape_name = "garagh-small-a"
        rate = 100

    class Garage_small_B(unittype.StaticType):
        id = "Garage small B"
        name = "Garage small B"
        shape_name = "garagh-small-b"
        rate = 100

    class Pump_station(unittype.StaticType):
        id = "Pump station"
        name = "Pump station"
        shape_name = "nasos"
        rate = 100

    class Oil_derrick(unittype.StaticType):
        id = "Oil derrick"
        name = "Oil derrick"
        shape_name = "neftevyshka"
        rate = 100

    class Container_red_1(unittype.StaticType):
        id = "Container red 1"
        name = "Container red 1"
        shape_name = "konteiner_red1"
        rate = 100

    class Container_red_2(unittype.StaticType):
        id = "Container red 2"
        name = "Container red 2"
        shape_name = "konteiner_red2"
        rate = 100

    class Container_red_3(unittype.StaticType):
        id = "Container red 3"
        name = "Container red 3"
        shape_name = "konteiner_red3"
        rate = 100

    class Container_white(unittype.StaticType):
        id = "Container white"
        name = "Container white"
        shape_name = "konteiner_white"
        rate = 100

    class Container_brown(unittype.StaticType):
        id = "Container brown"
        name = "Container brown"
        shape_name = "konteiner_brown"
        rate = 100

    class Barracks_2(unittype.StaticType):
        id = "Barracks 2"
        name = "Barracks 2"
        shape_name = "kazarma2"
        rate = 100

    class Military_staff(unittype.StaticType):
        id = "Military staff"
        name = "Military staff"
        shape_name = "aviashtab"
        rate = 100

    class Hangar_B(unittype.StaticType):
        id = "Hangar B"
        name = "Hangar B"
        shape_name = "angar_b"
        rate = 100

    class Fuel_tank(unittype.StaticType):
        id = "Fuel tank"
        name = "Fuel tank"
        shape_name = "toplivo-bak"
        rate = 100

    class Shelter_B(unittype.StaticType):
        id = "Shelter B"
        name = "Shelter B"
        shape_name = "ukrytie_b"
        rate = 100

    class Oil_platform(unittype.StaticType):
        id = "Oil platform"
        name = "Oil platform"
        shape_name = "plavbaza"
        rate = 100
        sea_object = True

    class Subsidiary_structure_1(unittype.StaticType):
        id = "Subsidiary structure 1"
        name = "Subsidiary structure 1"
        shape_name = "hozdomik1"
        rate = 100

    class Subsidiary_structure_2(unittype.StaticType):
        id = "Subsidiary structure 2"
        name = "Subsidiary structure 2"
        shape_name = "hozdomik2"
        rate = 100

    class Subsidiary_structure_3(unittype.StaticType):
        id = "Subsidiary structure 3"
        name = "Subsidiary structure 3"
        shape_name = "hozdomik3"
        rate = 100

    class Subsidiary_structure_A(unittype.StaticType):
        id = "Subsidiary structure A"
        name = "Subsidiary structure A"
        shape_name = "saray-a"
        rate = 100

    class Subsidiary_structure_B(unittype.StaticType):
        id = "Subsidiary structure B"
        name = "Subsidiary structure B"
        shape_name = "saray-b"
        rate = 100

    class Subsidiary_structure_C(unittype.StaticType):
        id = "Subsidiary structure C"
        name = "Subsidiary structure C"
        shape_name = "saray-c"
        rate = 100

    class Subsidiary_structure_D(unittype.StaticType):
        id = "Subsidiary structure D"
        name = "Subsidiary structure D"
        shape_name = "saray-d"
        rate = 100

    class Subsidiary_structure_E(unittype.StaticType):
        id = "Subsidiary structure E"
        name = "Subsidiary structure E"
        shape_name = "saray-e"
        rate = 100

    class Subsidiary_structure_F(unittype.StaticType):
        id = "Subsidiary structure F"
        name = "Subsidiary structure F"
        shape_name = "saray-f"
        rate = 100

    class Subsidiary_structure_G(unittype.StaticType):
        id = "Subsidiary structure G"
        name = "Subsidiary structure G"
        shape_name = "saray-g"
        rate = 100

    class Landmine(unittype.StaticType):
        id = "Landmine"
        name = "Landmine"
        shape_name = "landmine"
        rate = 1

    class FARP_Ammo_Storage(unittype.StaticType):
        id = "FARP Ammo Dump Coating"
        name = "FARP Ammo Storage"
        shape_name = "SetkaKP"
        rate = 50

    class FARP_Tent(unittype.StaticType):
        id = "FARP Tent"
        name = "FARP Tent"
        shape_name = "PalatkaB"
        rate = 50

    class FARP_Command_Post(unittype.StaticType):
        id = "FARP CP Blindage"
        name = "FARP Command Post"
        shape_name = "kp_ug"
        rate = 100

    class FARP_Fuel_Depot(unittype.StaticType):
        id = "FARP Fuel Depot"
        name = "FARP Fuel Depot"
        shape_name = "GSM Rus"
        rate = 20

    class GeneratorF(unittype.StaticType):
        id = "GeneratorF"
        name = "GeneratorF"
        shape_name = "GeneratorF"
        rate = 100

    class Airshow_cone(unittype.StaticType):
        id = "Airshow_Cone"
        name = "Airshow cone"
        shape_name = "Comp_cone"
        rate = 100

    class Airshow_Crowd(unittype.StaticType):
        id = "Airshow_Crowd"
        name = "Airshow Crowd"
        shape_name = "Crowd1"
        rate = 100

    class Coach_a_tank_yellow(unittype.StaticType):
        id = "Coach a tank yellow"
        name = "Coach a tank yellow"
        shape_name = "wagon-cisterna_yellow"
        rate = 100

    class Coach_a_tank_blue(unittype.StaticType):
        id = "Coach a tank blue"
        name = "Coach a tank blue"
        shape_name = "wagon-cisterna_blue"
        rate = 100

    class Coach_for_cargo(unittype.StaticType):
        id = "Coach cargo"
        name = "Coach for cargo"
        shape_name = "wagon-gruz"
        rate = 100

    class Coach_for_open_cargo(unittype.StaticType):
        id = "Coach cargo open"
        name = "Coach for open cargo"
        shape_name = "wagon-gruz-otkr"
        rate = 100

    class Coach_flatbed(unittype.StaticType):
        id = "Coach a platform"
        name = "Coach flatbed"
        shape_name = "wagon-platforma"
        rate = 100

    class Coach_for_passengers(unittype.StaticType):
        id = "Coach a passenger"
        name = "Coach for passengers"
        shape_name = "wagon-pass"
        rate = 100

    class Electric_locomotive(unittype.StaticType):
        id = "Electric locomotive"
        name = "Electric locomotive"
        shape_name = "elektrowoz"
        rate = 100

    class Locomotive(unittype.StaticType):
        id = "Locomotive"
        name = "Locomotive"
        shape_name = "teplowoz"
        rate = 100

    class Mark_Flag_Red(unittype.StaticType):
        id = "Red_Flag"
        name = "Mark Flag Red"
        shape_name = "H-flag_R"
        rate = 3

    class Mark_Flag_White(unittype.StaticType):
        id = "White_Flag"
        name = "Mark Flag White"
        shape_name = "H-Flag_W"
        rate = 3

    class Mark_Tyre_Black(unittype.StaticType):
        id = "Black_Tyre"
        name = "Mark Tyre Black"
        shape_name = "H-tyre_B"
        rate = 3

    class Mark_Tyre_White(unittype.StaticType):
        id = "White_Tyre"
        name = "Mark Tyre White"
        shape_name = "H-tyre_W"
        rate = 3

    class Mark_Tyre_with_Red_Flag(unittype.StaticType):
        id = "Black_Tyre_RF"
        name = "Mark Tyre with Red Flag"
        shape_name = "H-tyre_B_RF"
        rate = 3

    class Mark_Tyre_with_White_Flag(unittype.StaticType):
        id = "Black_Tyre_WF"
        name = "Mark Tyre with White Flag"
        shape_name = "H-tyre_B_WF"
        rate = 3

    class Windsock(unittype.StaticType):
        id = "Windsock"
        name = "Windsock"
        shape_name = "H-Windsock_RW"
        rate = 3

fortification_map = {
    ".Command Center": Fortification.Command_Center,
    "Hangar A": Fortification.Hangar_A,
    "Tech hangar A": Fortification.Tech_hangar_A,
    "Farm A": Fortification.Farm_A,
    "Farm B": Fortification.Farm_B,
    "Garage A": Fortification.Garage_A,
    "Boiler-house A": Fortification.Boiler_house_A,
    "Restaurant 1": Fortification.Restaurant_1,
    "Comms tower M": Fortification.Comms_tower_M,
    "Cafe": Fortification.Cafe,
    "Workshop A": Fortification.Workshop_A,
    "Electric power box": Fortification.Electric_power_box,
    "Supermarket A": Fortification.Supermarket_A,
    "Water tower A": Fortification.Water_tower_A,
    "TV tower": Fortification.TV_tower,
    "Shelter": Fortification.Shelter,
    "Repair workshop": Fortification.Repair_workshop,
    "Railway station": Fortification.Railway_station,
    "Railway crossing A": Fortification.Railway_crossing_A,
    "Railway crossing B": Fortification.Railway_crossing_B,
    "WC": Fortification.WC,
    "Small house 1A area": Fortification.Small_house_1A_area,
    "Small house 1A": Fortification.Small_house_1A,
    "Small house 1B area": Fortification.Small_house_1B_area,
    "Small house 1B": Fortification.Small_house_1B,
    "Small house 1C area": Fortification.Small_house_1C_area,
    "Small house 2C": Fortification.Small_house_2C,
    "Shop": Fortification.Shop,
    "Tech combine": Fortification.Tech_combine,
    "Chemical tank A": Fortification.Chemical_tank_A,
    "Small werehouse 1": Fortification.Small_warehouse_1,
    "Small werehouse 2": Fortification.Small_warehouse_2,
    "Small werehouse 3": Fortification.Small_warehouse_3,
    "Small werehouse 4": Fortification.Small_warehouse_4,
    "Garage B": Fortification.Garage_B,
    "Garage small A": Fortification.Garage_small_A,
    "Garage small B": Fortification.Garage_small_B,
    "Pump station": Fortification.Pump_station,
    "Oil derrick": Fortification.Oil_derrick,
    "Container red 1": Fortification.Container_red_1,
    "Container red 2": Fortification.Container_red_2,
    "Container red 3": Fortification.Container_red_3,
    "Container white": Fortification.Container_white,
    "Container brown": Fortification.Container_brown,
    "Barracks 2": Fortification.Barracks_2,
    "Military staff": Fortification.Military_staff,
    "Hangar B": Fortification.Hangar_B,
    "Fuel tank": Fortification.Fuel_tank,
    "Shelter B": Fortification.Shelter_B,
    "Oil platform": Fortification.Oil_platform,
    "Subsidiary structure 1": Fortification.Subsidiary_structure_1,
    "Subsidiary structure 2": Fortification.Subsidiary_structure_2,
    "Subsidiary structure 3": Fortification.Subsidiary_structure_3,
    "Subsidiary structure A": Fortification.Subsidiary_structure_A,
    "Subsidiary structure B": Fortification.Subsidiary_structure_B,
    "Subsidiary structure C": Fortification.Subsidiary_structure_C,
    "Subsidiary structure D": Fortification.Subsidiary_structure_D,
    "Subsidiary structure E": Fortification.Subsidiary_structure_E,
    "Subsidiary structure F": Fortification.Subsidiary_structure_F,
    "Subsidiary structure G": Fortification.Subsidiary_structure_G,
    "Landmine": Fortification.Landmine,
    "FARP Ammo Dump Coating": Fortification.FARP_Ammo_Storage,
    "FARP Tent": Fortification.FARP_Tent,
    "FARP CP Blindage": Fortification.FARP_Command_Post,
    "FARP Fuel Depot": Fortification.FARP_Fuel_Depot,
    "GeneratorF": Fortification.GeneratorF,
    "Airshow_Cone": Fortification.Airshow_cone,
    "Airshow_Crowd": Fortification.Airshow_Crowd,
    "Coach a tank yellow": Fortification.Coach_a_tank_yellow,
    "Coach a tank blue": Fortification.Coach_a_tank_blue,
    "Coach cargo": Fortification.Coach_for_cargo,
    "Coach cargo open": Fortification.Coach_for_open_cargo,
    "Coach a platform": Fortification.Coach_flatbed,
    "Coach a passenger": Fortification.Coach_for_passengers,
    "Electric locomotive": Fortification.Electric_locomotive,
    "Locomotive": Fortification.Locomotive,
    "Red_Flag": Fortification.Mark_Flag_Red,
    "White_Flag": Fortification.Mark_Flag_White,
    "Black_Tyre": Fortification.Mark_Tyre_Black,
    "White_Tyre": Fortification.Mark_Tyre_White,
    "Black_Tyre_RF": Fortification.Mark_Tyre_with_Red_Flag,
    "Black_Tyre_WF": Fortification.Mark_Tyre_with_White_Flag,
    "Windsock": Fortification.Windsock,
}


class GroundObject:

    class Building(unittype.StaticType):
        id = "Building"
        name = "Building"
        category = ""

    class Bridge(unittype.StaticType):
        id = "Bridge"
        name = "Bridge"
        category = ""

    class Transport(unittype.StaticType):
        id = "Transport"
        name = "Transport"
        category = ""

    class Train(unittype.StaticType):
        id = "Train"
        name = "Train"
        category = ""

groundobject_map = {
    "Building": GroundObject.Building,
    "Bridge": GroundObject.Bridge,
    "Transport": GroundObject.Transport,
    "Train": GroundObject.Train,
}


class Warehouse:

    class Warehouse(unittype.StaticType):
        id = "Warehouse"
        name = "Warehouse"
        shape_name = "sklad"
        category = "Warehouses"
        rate = 100

    class Tank(unittype.StaticType):
        id = "Tank"
        name = "Tank"
        shape_name = "bak"
        category = "Warehouses"
        rate = 100

    class Ammunition_depot(unittype.StaticType):
        id = ".Ammunition depot"
        name = "Ammunition depot"
        shape_name = "SkladC"
        category = "Warehouses"
        rate = 100

    class Tank_2(unittype.StaticType):
        id = "Tank 2"
        name = "Tank 2"
        shape_name = "airbase_tbilisi_tank_01"
        category = "Warehouses"
        rate = 100

    class Tank_3(unittype.StaticType):
        id = "Tank 3"
        name = "Tank 3"
        shape_name = "airbase_tbilisi_tank_02"
        category = "Warehouses"
        rate = 100

warehouse_map = {
    "Warehouse": Warehouse.Warehouse,
    "Tank": Warehouse.Tank,
    ".Ammunition depot": Warehouse.Ammunition_depot,
    "Tank 2": Warehouse.Tank_2,
    "Tank 3": Warehouse.Tank_3,
}


class Cargo:

    class uh1h_cargo(unittype.StaticType):
        id = "uh1h_cargo"
        name = "uh1h_cargo"
        shape_name = "ab-212_cargo"
        category = "Cargos"
        rate = 100
        can_cargo = True

    class ammo_cargo(unittype.StaticType):
        id = "ammo_cargo"
        name = "ammo_cargo"
        shape_name = "ammo_box_cargo"
        category = "Cargos"
        rate = 1000
        can_cargo = True

    class f_bar_cargo(unittype.StaticType):
        id = "f_bar_cargo"
        name = "f_bar_cargo"
        shape_name = "f_bar_cargo"
        category = "Cargos"
        rate = 823
        can_cargo = True

    class m117_cargo(unittype.StaticType):
        id = "m117_cargo"
        name = "m117_cargo"
        shape_name = "m117_cargo"
        category = "Cargos"
        rate = 840
        can_cargo = True

    class iso_container(unittype.StaticType):
        id = "iso_container"
        name = "iso_container"
        shape_name = "iso_container_cargo"
        category = "Cargos"
        rate = 3800
        can_cargo = True

    class iso_container_small(unittype.StaticType):
        id = "iso_container_small"
        name = "iso_container_small"
        shape_name = "iso_container_small_cargo"
        category = "Cargos"
        rate = 220
        can_cargo = True

    class barrels_cargo(unittype.StaticType):
        id = "barrels_cargo"
        name = "barrels_cargo"
        shape_name = "barrels_cargo"
        category = "Cargos"
        rate = 480
        can_cargo = True

    class container_cargo(unittype.StaticType):
        id = "container_cargo"
        name = "container_cargo"
        shape_name = "bw_container_cargo"
        category = "Cargos"
        rate = 100
        can_cargo = True

    class tetrapod_cargo(unittype.StaticType):
        id = "tetrapod_cargo"
        name = "tetrapod_cargo"
        shape_name = "tetrapod_cargo"
        category = "Cargos"
        rate = 5000
        can_cargo = True

    class fueltank_cargo(unittype.StaticType):
        id = "fueltank_cargo"
        name = "fueltank_cargo"
        shape_name = "fueltank_cargo"
        category = "Cargos"
        rate = 800
        can_cargo = True

    class oiltank_cargo(unittype.StaticType):
        id = "oiltank_cargo"
        name = "oiltank_cargo"
        shape_name = "oiltank_cargo"
        category = "Cargos"
        rate = 700
        can_cargo = True

    class pipes_big_cargo(unittype.StaticType):
        id = "pipes_big_cargo"
        name = "pipes_big_cargo"
        shape_name = "pipes_big_cargo"
        category = "Cargos"
        rate = 4815
        can_cargo = True

    class pipes_small_cargo(unittype.StaticType):
        id = "pipes_small_cargo"
        name = "pipes_small_cargo"
        shape_name = "pipes_small_cargo"
        category = "Cargos"
        rate = 4350
        can_cargo = True

    class trunks_small_cargo(unittype.StaticType):
        id = "trunks_small_cargo"
        name = "trunks_small_cargo"
        shape_name = "trunks_small_cargo"
        category = "Cargos"
        rate = 5000
        can_cargo = True

    class trunks_long_cargo(unittype.StaticType):
        id = "trunks_long_cargo"
        name = "trunks_long_cargo"
        shape_name = "trunks_long_cargo"
        category = "Cargos"
        rate = 4747
        can_cargo = True

cargo_map = {
    "uh1h_cargo": Cargo.uh1h_cargo,
    "ammo_cargo": Cargo.ammo_cargo,
    "f_bar_cargo": Cargo.f_bar_cargo,
    "m117_cargo": Cargo.m117_cargo,
    "iso_container": Cargo.iso_container,
    "iso_container_small": Cargo.iso_container_small,
    "barrels_cargo": Cargo.barrels_cargo,
    "container_cargo": Cargo.container_cargo,
    "tetrapod_cargo": Cargo.tetrapod_cargo,
    "fueltank_cargo": Cargo.fueltank_cargo,
    "oiltank_cargo": Cargo.oiltank_cargo,
    "pipes_big_cargo": Cargo.pipes_big_cargo,
    "pipes_small_cargo": Cargo.pipes_small_cargo,
    "trunks_small_cargo": Cargo.trunks_small_cargo,
    "trunks_long_cargo": Cargo.trunks_long_cargo,
}



import copy
from typing import List
from enum import Enum

from dcs import mapping
from dcs import action
from dcs import condition


class TriggerZone:
    def __init__(self, _id, position: mapping.Point, radius=1500, hidden=False, name=""):
        self.id = _id
        self.radius = radius
        self.position = copy.copy(position)
        self.hidden = hidden
        self.name = name
        self.color = {1: 1, 2: 1, 3: 1, 4: 0.15}  # TODO color attributes

    def dict(self):
        return {
            "name": self.name,
            "hidden": self.hidden,
            "x": self.position.x,
            "y": self.position.y,
            "zoneId": self.id,
            "radius": self.radius,
            "color": self.color
        }

    def __repr__(self):
        return "TriggerZone({id}, {x}, {y}, {r}, '{n}')".format(
            id=self.id, x=self.position.x, y=self.position.y, r=self.radius, n=self.name
        )


class Triggers:
    def __init__(self):
        self.current_zone_id = 0
        self._zones = []  # type: List[TriggerZone]

    def load_from_dict(self, data):
        self.current_zone_id = 0
        self._zones = []
        for x in data["zones"]:
            imp_zone = data["zones"][x]
            tz = TriggerZone(
                imp_zone["zoneId"],
                mapping.Point(imp_zone["x"], imp_zone["y"]),
                imp_zone["radius"],
                imp_zone["hidden"],
                imp_zone["name"]
            )
            tz.color = imp_zone["color"]
            self._zones.append(tz)
            self.current_zone_id = max(self.current_zone_id, tz.id)

    def add_triggerzone(self, position: mapping.Point, radius=1500, hidden=False, name="") -> TriggerZone:
        self.current_zone_id += 1
        tz = TriggerZone(self.current_zone_id, position, radius, hidden, name)
        self._zones.append(tz)
        return tz

    def zones(self) -> List[TriggerZone]:
        return self._zones

    def dict(self):
        return {
            "zones": {i + 1: self._zones[i].dict() for i in range(0, len(self._zones))}
        }


class Event(Enum):
    NoEvent = ""
    Destroy = "dead"
    Shot = "shot"
    Crash = "crash"
    Eject = "eject"
    Refuel = "refuel"
    PilotDead = "pilot dead"
    BaseCaptured = "base captured"
    TookControl = "took control"
    RefuelStop = "refuel stop"
    Failure = "failure"


class TriggerRule:
    predicate = None

    def __init__(self, event, comment):
        self.comment = comment
        self.eventlist = event
        self.rules = []  # type: List[condition.Condition]
        self.actions = []  # type: List[action.Action]

    @classmethod
    def create_from_dict(cls, mission, d):
        trig = cls()
        trig.comment = d["comment"]
        trig.event = Event(d["eventlist"])
        actions = d["actions"]
        for a in actions:
            action_ = action.actions_map[actions[a]["predicate"]].create_from_dict(actions[a], mission)
            trig.actions.append(action_)
        rules = d["rules"]
        for r in rules:
            rule = condition.condition_map[rules[r]["predicate"]].create_from_dict(rules[r])
            trig.rules.append(rule)
        return trig

    def condition_str(self):
        if self.rules:
            return "return(" + " and ".join(map(repr, self.rules)) + ")"
        return "return(true)"

    def action_str(self, idx):
        actionstr = "; ".join([repr(x) for x in self.actions])
        if self.eventlist != Event.NoEvent:
            actionstr += '; mission.trig.events["' + self.eventlist.value + '"][' + str(idx) + ']=nil;'
        else:
            actionstr += '; mission.trig.func[' + str(idx) + ']=nil;'
        return actionstr

    def func_str(self, start, idx):
        if self.eventlist == Event.NoEvent and (start and isinstance(self, TriggerStart) or
                                                (not start and not isinstance(self, TriggerStart))):
            if isinstance(self, TriggerCondition):
                return "if mission.trig.conditions[{idx}]() then if not mission.trig.flag[{idx}" \
                       "] then mission.trig.actions[{idx}](); mission.trig.flag[{idx}" \
                       "] = true;end; else mission.trig.flag[{idx}] = false; end;".format(idx=idx)
            else:
                return "if mission.trig.conditions[{idx}]() then mission.trig.actions[{idx}]() end".format(idx=idx)
        return None

    def events_str(self, event, idx):
        if self.eventlist == event:
            return "if mission.trig.conditions[{idx}]() then mission.trig.actions[{idx}]() end".format(idx=idx)

    def dict(self):
        return {
            "comment": self.comment,
            "eventlist": self.eventlist.value,
            "predicate": self.predicate,
            "rules": {i + 1: self.rules[i].dict() for i in range(0, len(self.rules))},
            "actions": {i + 1: self.actions[i].dict() for i in range(0, len(self.actions))}
        }

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self.dict()) + ')'


class TriggerOnce(TriggerRule):
    predicate = "triggerOnce"

    def __init__(self, event: Event = Event.NoEvent, comment=""):
        super(TriggerOnce, self).__init__(event, comment)


class TriggerContinious(TriggerRule):
    predicate = "triggerContinious"

    def __init__(self, event: Event = Event.NoEvent, comment=""):
        super(TriggerContinious, self).__init__(event, comment)


class TriggerStart(TriggerRule):
    predicate = "triggerStart"

    def __init__(self, comment=""):
        super(TriggerStart, self).__init__(Event.NoEvent, comment)


class TriggerCondition(TriggerRule):
    predicate = "triggerFront"

    def __init__(self, comment=""):
        super(TriggerCondition, self).__init__(Event.NoEvent, comment)

trigger_map = {
    TriggerOnce.predicate: TriggerOnce,
    TriggerContinious.predicate: TriggerContinious,
    TriggerCondition.predicate: TriggerCondition,
    TriggerStart.predicate: TriggerStart
}


class Rules:
    def __init__(self):
        self.triggers = []  # type: List[TriggerRule]

    def load_from_dict(self, mission, d):
        self.triggers.clear()
        for x in d:
            self.triggers.append(trigger_map[d[x]["predicate"]].create_from_dict(mission, d[x]))

    def trig(self):
        d = {}
        d["conditions"] = {i + 1: self.triggers[i].condition_str() for i in range(0, len(self.triggers))}
        d["actions"] = {i + 1: self.triggers[i].action_str(i + 1) for i in range(0, len(self.triggers))}
        d["func"] = {i + 1: self.triggers[i].func_str(False, i + 1) for i in range(0, len(self.triggers))
                     if self.triggers[i].func_str(False, i + 1)}
        d["funcStartup"] = {i + 1: self.triggers[i].func_str(True, i + 1) for i in range(0, len(self.triggers))
                            if self.triggers[i].func_str(True, i + 1)}
        d["customStartup"] = {}
        d["events"] = {}
        for e in Event:
            if e != Event.NoEvent:
                events = {i + 1: self.triggers[i].events_str(e, i + 1)
                          for i in range(0, len(self.triggers))
                          if self.triggers[i].events_str(e, i + 1)}
                if events:
                    d["events"][e.value] = events

        d["flag"] = {i + 1: True for i in range(0, len(self.triggers))}
        return d

    def trigrules(self):
        return {i + 1: self.triggers[i].dict() for i in range(0, len(self.triggers))}



import unittest
from dcs.lua.parse import loads


class TestNumber(unittest.TestCase):

    def test_integer(self):
        r = loads("int = 3")
        self.assertEqual(r, {"int": 3})

        r = loads("int = 193984")
        self.assertEqual(r, {"int": 193984})

        r = loads("int = -23")
        self.assertEqual(r, {"int": -23})

    def test_decimal(self):
        r = loads("dec = 666.6")
        self.assertEqual(r, {"dec": 666.6})

    def test_string(self):
        s = """
mission =
{
    ["trig"] =
    {
        ["actions"] =
        {
            [1] = "a_(get(\\"ResKey_Action_62\\")); mission.trig.func[1]=nil;",
            [2] = "a_do_script_file(getValueResourceByKey(\\"ResKey_Action_64\\")); mission.trig.func[2]=nil;",
            [3] = "a_out_text_delay(getValueDictByKey(\\"DictKey_ActionText_66\\"), 10, false);a_do_script(getValueDictByKey(\\"DictKey_ActionText_255\\")); mission.trig.func[3]=nil;",
            [4] = "a_add_radio_item(getValueDictByKey(\\"Destroy Red Ground Group\\"), 1500, 1); mission.trig.func[4]=nil;",
            [5] = "a_do_script(getValueDictByKey(\\"DictKey_ActionText_213\\"));a_clear_flag(1500);",
        }, -- end of ["actions"]
    }
}
"""
        r = loads(s)

        r = loads('x = "a_do_script_file(getValueResourceByKey(\\"ResKey_Action_62\\")); mission.trig.func[1]=nil;"')
        self.assertEqual(r, {"x": "a_do_script_file(getValueResourceByKey(\"ResKey_Action_62\")); mission.trig.func[1]=nil;"})

    def test_object(self):
        luas = """
mission=
{
    ["trig"]=
    {
    },

    ["coalitions"]=
    {
        ["blue"]=
        {
            [1]=11,
            [2]=4,
            [3]=6
        }
    }, -- end of ["coalitions"]
    ["maxDictId"]=18,
    ["descriptionBlueTask"]="DictKey_descriptionBlueTask_3"
}
"""
        ref = {'mission': {
            'coalitions': {'blue': {1: 11.0, 2: 4.0, 3: 6.0}},
            'descriptionBlueTask': 'DictKey_descriptionBlueTask_3',
            'trig': {},
            'maxDictId': 18.0}}
        r = loads(luas)
        self.assertEqual(r, ref)

    def test_dictmix(self):
        luas = """
o =
{
    ["callsign"] =
    {
        [1] = 1,
        [2] = 1,
        [3] = 1,
        ["name"] = "Enfield11",
    } -- end of ["callsign"]
}
"""
        ref = {"o": {
            'callsign': {
                1: 1,
                2: 1,
                3: 1,
                "name": "Enfield11"
            }
        }}
        r = loads(luas)
        self.assertEqual(r, ref)

    def test_syntaxerr(self):
        with self.assertRaises(SyntaxError):
            loads("""m=
            {
                ["x"] 12
            }""")


if __name__ == '__main__':
    unittest.main()



# This file is generated from pydcs_export.lua

from . import unittype


class Armed_speedboat(unittype.ShipType):
    id = "speedboat"
    name = "Armed speedboat"


class CVN_70_Carl_Vinson(unittype.ShipType):
    id = "VINSON"
    name = "CVN-70 Carl Vinson"
    plane_num = 72
    helicopter_num = 6
    parking = 4


class FFG_7CL_Oliver_Hazzard_Perry(unittype.ShipType):
    id = "PERRY"
    name = "FFG-7CL Oliver Hazzard Perry"
    helicopter_num = 2
    parking = 1


class CG_60_Normandy(unittype.ShipType):
    id = "TICONDEROG"
    name = "CG-60 Normandy"
    helicopter_num = 2
    parking = 1


class FFL_1124_4_Grisha(unittype.ShipType):
    id = "ALBATROS"
    name = "FFL 1124.4 Grisha"


class CV_1143_5_Admiral_Kuznetsov(unittype.ShipType):
    id = "KUZNECOW"
    name = "CV 1143.5 Admiral Kuznetsov"
    plane_num = 24
    helicopter_num = 12
    parking = 3


class FSG_1241_1MP_Molniya(unittype.ShipType):
    id = "MOLNIYA"
    name = "FSG 1241.1MP Molniya"


class CG_1164_Moskva(unittype.ShipType):
    id = "MOSCOW"
    name = "CG 1164 Moskva"
    helicopter_num = 1
    parking = 1


class FFG_11540_Neustrashimy(unittype.ShipType):
    id = "NEUSTRASH"
    name = "FFG 11540 Neustrashimy"
    helicopter_num = 1
    parking = 1


class CGN_1144_2_Pyotr_Velikiy(unittype.ShipType):
    id = "PIOTR"
    name = "CGN 1144.2 Pyotr Velikiy"
    helicopter_num = 1
    parking = 1


class FF_1135M_Rezky(unittype.ShipType):
    id = "REZKY"
    name = "FF 1135M Rezky"


class Tanker_Elnya_160(unittype.ShipType):
    id = "ELNYA"
    name = "Tanker Elnya 160"


class Dry_cargo_ship_Ivanov(unittype.ShipType):
    id = "Dry-cargo ship-2"
    name = "Dry cargo ship Ivanov"


class Bulk_cargo_ship_Yakushev(unittype.ShipType):
    id = "Dry-cargo ship-1"
    name = "Bulk cargo ship Yakushev"


class Civil_boat_Zvezdny(unittype.ShipType):
    id = "ZWEZDNY"
    name = "Civil boat Zvezdny"


class SSK_877(unittype.ShipType):
    id = "KILO"
    name = "SSK 877"


class SSK_641B(unittype.ShipType):
    id = "SOM"
    name = "SSK 641B"

ship_map = {
    "speedboat": Armed_speedboat,
    "VINSON": CVN_70_Carl_Vinson,
    "PERRY": FFG_7CL_Oliver_Hazzard_Perry,
    "TICONDEROG": CG_60_Normandy,
    "ALBATROS": FFL_1124_4_Grisha,
    "KUZNECOW": CV_1143_5_Admiral_Kuznetsov,
    "MOLNIYA": FSG_1241_1MP_Molniya,
    "MOSCOW": CG_1164_Moskva,
    "NEUSTRASH": FFG_11540_Neustrashimy,
    "PIOTR": CGN_1144_2_Pyotr_Velikiy,
    "REZKY": FF_1135M_Rezky,
    "ELNYA": Tanker_Elnya_160,
    "Dry-cargo ship-2": Dry_cargo_ship_Ivanov,
    "Dry-cargo ship-1": Bulk_cargo_ship_Yakushev,
    "ZWEZDNY": Civil_boat_Zvezdny,
    "KILO": SSK_877,
    "SOM": SSK_641B,
}



from .countries import Russia, USA
from . import unit
from .mission import Mission
from . import mapping
import dcs.vehicles


class VehicleTemplate:
    class Russia:
        @staticmethod
        def sa10_site(mission: Mission, position: mapping.Point, heading, prefix="", skill=unit.Skill.Average):
            russia = mission.country("Russia")
            vg = mission.vehicle_group(russia, prefix + "SA10 site",
                                       Russia.Vehicle.AirDefence.SAM_SA_10_S_300PS_CP_54K6, position, heading)
            u = mission.vehicle("Operator 1", Russia.Vehicle.Infantry.Infantry_Soldier_Rus)
            u.position = position.point_from_heading(heading + 180, 10)
            u.heading = heading
            vg.add_unit(u)

            hdg = 90
            for i in range(0, 3):  # 3 launchers
                u = mission.vehicle("launcher #" + str(i+1), Russia.Vehicle.AirDefence.SAM_SA_10_S_300PS_LN_5P85C)
                u.position = position.point_from_heading(heading + hdg, 50)
                u.heading = heading
                vg.add_unit(u)
                hdg += 90

            u = mission.vehicle("radar", Russia.Vehicle.AirDefence.SAM_SA_10_S_300PS_TR_30N6)
            u.position = position.point_from_heading(heading, 80)
            u.heading = heading
            vg.add_unit(u)

            u = mission.vehicle("radar", Russia.Vehicle.AirDefence.SAM_SA_10_S_300PS_SR_64H6E)
            u.position = position.point_from_heading(heading + 180, 100)
            u.heading = heading
            vg.add_unit(u)

            for u in vg.units:
                u.skill = skill

            return vg

    class USA:
        @staticmethod
        def patriot_site(mission: Mission, position, heading, prefix="", skill=unit.Skill.Average):
            usa = mission.country("USA")
            vg = mission.vehicle_group(usa, prefix + "Patriot site", USA.Vehicle.AirDefence.SAM_Patriot_ICC, position, heading)
            u = mission.vehicle("Operator 1", USA.Vehicle.Infantry.Infantry_M4)
            u.position = position.point_from_heading(heading + 180, 5)
            u.heading = heading
            vg.add_unit(u)

            hdg = 90
            for i in range(0, 2):  # 2 launchers
                u = mission.vehicle("launcher #" + str(i+1), USA.Vehicle.AirDefence.SAM_Patriot_LN_M901)
                u.position = position.point_from_heading(heading + hdg, 50)
                u.heading = heading
                vg.add_unit(u)
                hdg += 90

            u = mission.vehicle("Electronic power plant", USA.Vehicle.AirDefence.SAM_Patriot_EPP_III)
            u.position = position.point_from_heading(heading + 180, 50)
            u.heading = heading
            vg.add_unit(u)

            u = mission.vehicle("radar", USA.Vehicle.AirDefence.SAM_Patriot_STR_AN_MPQ_53)
            u.position = position.point_from_heading(heading, 80)
            u.heading = heading
            vg.add_unit(u)

            inf = mission.vehicle("Operator 2", USA.Vehicle.Infantry.Infantry_M4)
            inf.position = position.point_from_heading(heading + 270, 5)
            vg.add_unit(inf)

            u = mission.vehicle("Antenna", USA.Vehicle.AirDefence.SAM_Patriot_AMG_AN_MRC_137)
            u.position = position.point_from_heading(heading + 180, 100)
            u.heading = heading
            vg.add_unit(u)

            u = mission.vehicle("ECS", USA.Vehicle.AirDefence.SAM_Patriot_ECS_AN_MSQ_104)
            u.position = position.point_from_heading(heading + 120, 80)
            u.heading = heading
            vg.add_unit(u)

            for u in vg.units:
                u.skill = skill

        @staticmethod
        def hawk_site(mission: Mission, position, heading, prefix="", skill=unit.Skill.Average):
            usa = mission.country("USA")
            vg = mission.vehicle_group(usa, prefix + "Hawk site", USA.Vehicle.AirDefence.SAM_Hawk_PCP, position, heading)

            u = mission.vehicle("Operator 1", USA.Vehicle.Infantry.Infantry_M4)
            u.position = position.point_from_heading(heading + 180, 5)
            u.heading = heading
            vg.add_unit(u)

            hdg = 90
            for i in range(0, 2):  # 2 launchers
                u = mission.vehicle("launcher #" + str(i+1), USA.Vehicle.AirDefence.SAM_Hawk_LN_M192)
                u.position = position.point_from_heading(heading + hdg, 50)
                u.heading = heading
                vg.add_unit(u)
                hdg += 90

            u = mission.vehicle("Radar", USA.Vehicle.AirDefence.SAM_Hawk_SR_AN_MPQ_50)
            u.position = position.point_from_heading(heading + 180, 20)
            u.heading = heading
            vg.add_unit(u)

            inf = mission.vehicle("Operator 2", USA.Vehicle.Infantry.Infantry_M4)
            inf.position = position.point_from_heading(heading + 270, 5)
            vg.add_unit(inf)

            u = mission.vehicle("Tower", USA.Vehicle.AirDefence.SAM_Hawk_TR_AN_MPQ_46)
            u.position = position.point_from_heading(heading + 80, 80)
            u.heading = heading
            vg.add_unit(u)

            u = mission.vehicle("Wave Radar", USA.Vehicle.AirDefence.SAM_Hawk_CWAR_AN_MPQ_55)
            u.position = position.point_from_heading(heading + 180, 100)
            u.heading = heading
            vg.add_unit(u)

            for u in vg.units:
                u.skill = skill

    @staticmethod
    def sa11_site(mission, country, position, heading, prefix="", skill=unit.Skill.Average):
        vg = mission.vehicle_group(country, prefix + "SA10 site",
                                   dcs.vehicles.AirDefence.SAM_SA_11_Buk_CC_9S470M1, position, heading)

        u = mission.vehicle("Operator 1", Russia.Vehicle.Infantry.Infantry_Soldier_Rus)
        u.position = position.point_from_heading(heading + 180, 10)
        u.heading = heading
        vg.add_unit(u)

        hdg = 90
        for i in range(0, 2):  # 2 launchers
            u = mission.vehicle("launcher #" + str(i + 1), dcs.vehicles.AirDefence.SAM_SA_11_Buk_LN_9A310M1)
            u.position = position.point_from_heading(heading + hdg, 50)
            u.heading = heading
            vg.add_unit(u)
            hdg += 90

        u = mission.vehicle("radar", dcs.vehicles.AirDefence.SAM_SA_11_Buk_SR_9S18M1)
        u.position = position.point_from_heading(heading, 80)
        u.heading = heading
        vg.add_unit(u)

        for u in vg.units:
            u.skill = skill

        return vg



# terrain module
from typing import List, Dict, Optional, Tuple, Set
from collections import defaultdict, deque
from dcs import mapping, lua, point
from dcs import unittype
import random
import pickle


class ParkingSlot:
    def __init__(self,
                 crossroad_idx,
                 position: mapping.Point,
                 large=False,
                 slot_name=None,
                 heli=False,
                 airplanes=True,
                 length=None,
                 width=None,
                 height=None,
                 shelter=False):
        self.crossroad_idx = crossroad_idx
        self.position = position
        self.length = length
        self.width = width
        self.height = height
        self.helicopter = heli
        self.airplanes = airplanes
        self.large = large
        self.shelter = shelter
        self.unit_id = None  # type: int
        self.slot_name = slot_name

    def __repr__(self):
        return 'ParkingSlot({id}, "{name}", large={large}, heli={heli})'.format(
            id=self.crossroad_idx, name=self.slot_name, large=self.large, heli=self.helicopter
        )


class Runway:
    def __init__(self, heading, ils=None, leftright=0):
        """

        :param heading: Compass direction of runway
        :param ils:
        :param leftright: 0 only 1 runway
                          1 left runway
                          2 right runway
        :return: None
        """
        self.heading = heading
        self.ils = ils
        self.leftright = leftright


class RunwayOccupiedError(RuntimeError):
    pass


class NoParkingSlotError(RuntimeError):
    pass


class Airport:
    id = None
    name = None
    position = None  # type: mapping.Point
    tacan = None
    frequencies = []
    unit_zones = []  # type: List[mapping.Rectangle]
    civilian = True
    slot_version = 1

    def __init__(self):
        self.runway_used = None
        self.runways = []  # type: List[Runway]
        self.parking_slots = []  # type: List[ParkingSlot]

        # warehouse values
        self.coalition = "NEUTRAL"
        self.size = 100
        self.speed = 16.666666
        self.periodicity = 30
        self.unlimited_munitions = True
        self.unlimited_aircrafts = True
        self.unlimited_fuel = True
        self.operating_level_air = 10
        self.operating_level_fuel = 10
        self.operating_level_equipment = 10
        self.aircrafts = {}
        self.weapons = {}
        self.suppliers = {}
        self.gasoline_init = 100
        self.methanol_mixture_init = 100
        self.diesel_init = 100
        self.jet_init = 100

    def load_from_dict(self, d):
        self.coalition = d["coalition"]
        self.speed = d["speed"]
        self.size = d["size"]
        self.periodicity = d["periodicity"]
        self.unlimited_munitions = d["unlimitedMunitions"]
        self.unlimited_fuel = d["unlimitedFuel"]
        self.unlimited_aircrafts = d["unlimitedAircrafts"]
        self.operating_level_air = d["OperatingLevel_Air"]
        self.operating_level_equipment = d["OperatingLevel_Eqp"]
        self.operating_level_fuel = d["OperatingLevel_Fuel"]
        self.gasoline_init = d.get("gasoline", {}).get("InitFuel", 100)
        self.diesel_init = d.get("diesel", {}).get("InitFuel", 100)
        self.jet_init = d.get("jet_fuel", {}).get("InitFuel", 100)
        self.methanol_mixture_init = d.get("methanol_mixture", {}).get("InitFuel", 100)
        self.aircrafts = d["aircrafts"]
        self.weapons = d["weapons"]

    def set_blue(self):
        self.set_coalition("BLUE")

    def set_red(self):
        self.set_coalition("RED")

    def set_neutral(self):
        self.set_coalition("NEUTRAL")

    def set_coalition(self, side):
        if side.lower() in ["red", "blue", "neutral"]:
            self.coalition = side.upper()
            return True
        return False

    def is_red(self):
        return self.coalition == "RED"

    def is_blue(self):
        return self.coalition == "BLUE"

    def random_unit_zone(self) -> mapping.Rectangle:
        if self.unit_zones:
            return self.unit_zones[random.randrange(0, len(self.unit_zones))]
        return mapping.Rectangle.from_point(mapping.Point(self.position.x + 500, self.position.y), 200)

    def occupy_runway(self, group):
        if self.runway_used is not None:
            raise RunwayOccupiedError("Runway already in use by group {gname}".format(gname=self.runway_used.name))
        self.runway_used = group
        return self.runway_used

    def parking_slot(self, index: int) -> ParkingSlot:
        """Searches the parking slot with the given crossroad index.

        Args:
            index: crossroad index of the parkings slot

        Returns:
            ParkingSlot: the found slot or None if not found.
        """
        for x in self.parking_slots:
            if x.crossroad_idx == index:
                return x
        return None

    def _free_parking_slots_resolve_v1(self, large: bool, helicopter: bool) -> List[ParkingSlot]:
        slots_index = range(0, len(self.parking_slots))
        free_large_slots = {x for x in slots_index
                            if self.parking_slots[x].unit_id is None and
                            self.parking_slots[x].large}

        large_slots = sorted([self.parking_slots[x] for x in free_large_slots], key=lambda x: x.slot_name)
        if large:
            return large_slots

        free_heli_slots = {x for x in slots_index
                           if self.parking_slots[x].unit_id is None and
                           self.parking_slots[x].helicopter and not self.parking_slots[x].large}

        heli_slots = sorted([self.parking_slots[x] for x in free_heli_slots], key=lambda x: x.slot_name)
        if helicopter:
            return heli_slots + large_slots

        free_slots = list(self.parking_slots[x] for x in slots_index if self.parking_slots[x].unit_id is None and
                          not self.parking_slots[x].large and not self.parking_slots[x].helicopter)
        free_slots = sorted(free_slots, key=lambda x: x.slot_name)
        return free_slots + heli_slots + large_slots

    def _free_parking_slots_resolve_v2(self, aircraft_type: unittype.FlyingType) -> List[ParkingSlot]:
        free_slots = [x for x in self.parking_slots
                      if x.unit_id is None and
                      aircraft_type.width < x.width and
                      aircraft_type.height < (x.height or 1000) and
                      aircraft_type.length < x.length and
                      aircraft_type.helicopter == x.helicopter and
                      not aircraft_type.helicopter == x.airplanes]

        slots_sorted = sorted(free_slots, key=lambda x: x.slot_name)
        return slots_sorted

    def free_parking_slots(self, aircraft_type: unittype.FlyingType) -> List[ParkingSlot]:
        if self.slot_version == 1:
            return self._free_parking_slots_resolve_v1(aircraft_type.large_parking_slot, aircraft_type.helicopter)
        else:
            return self._free_parking_slots_resolve_v2(aircraft_type)

    def free_parking_slot(self, aircraft_type: unittype.FlyingType) -> Optional[ParkingSlot]:
        slots = self.free_parking_slots(aircraft_type)
        if slots:
            return slots[0]
        return None

    def dict(self):
        d = {
            "coalition": self.coalition,
            "speed": self.speed,
            "size": self.size,
            "periodicity": self.periodicity,
            "unlimitedMunitions": self.unlimited_munitions,
            "unlimitedFuel": self.unlimited_fuel,
            "unlimitedAircrafts": self.unlimited_aircrafts,
            "OperatingLevel_Air": self.operating_level_air,
            "OperatingLevel_Eqp": self.operating_level_equipment,
            "OperatingLevel_Fuel": self.operating_level_fuel,
            "gasoline": {"InitFuel": self.gasoline_init},
            "diesel": {"InitFuel": self.diesel_init},
            "jet_fuel": {"InitFuel": self.jet_init},
            "methanol_mixture": {"InitFuel": self.methanol_mixture_init},
            "aircrafts": self.aircrafts,
            "weapons": self.weapons,
            "suppliers": self.suppliers
        }
        return d

    def __repr__(self):
        return "Airport(" + self.name + ")"


class MapView:
    def __init__(self, center: mapping.Point, zoom=1000000):
        self.position = center
        self.zoom = zoom

    def load_from_dict(self, d):
        self.position = mapping.Point(d["centerX"], d["centerY"])
        self.zoom = d["zoom"]

    def dict(self):
        return {
            "centerX": self.position.x,
            "centerY": self.position.y,
            "zoom": self.zoom
        }


class Node:
    def __init__(self, name, rating, position: mapping.Point):
        self.name = name
        self.rating = rating
        self.position = position  # type: mapping.Point
        """Air defence positions for small arms, mostly on buildings"""
        self.air_defence_pos_small = []  # type: List[mapping.Point]

    def __repr__(self):
        return 'Node("{name}", {r})'.format(name=self.name, r=self.rating)


class Graph:
    Edge_indicators = {'N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW'}

    def __init__(self):
        self.nodes = set()
        self.edges = defaultdict(list)
        self.edge_properties = {}

    def node(self, node_name) -> Node:
        for x in self.nodes:
            if x.name == node_name:
                return x
        raise RuntimeError('Node not found: ' + node_name)

    def node_names(self) -> Set[str]:
        return {x.name for x in self.nodes}

    def rated_nodes(self, min_rating=0) -> Set[Node]:
        return {x for x in self.nodes if x.rating and x.rating > min_rating}

    def nodes_within(self, polygon: mapping.Polygon) -> List[Node]:
        return [x for x in self.nodes if polygon.point_in_poly(x.position)]

    def rated_nodes_within(self, polygon: mapping.Polygon, min_rating=0) -> List[Node]:
        return [x for x in self.rated_nodes(min_rating) if polygon.point_in_poly(x.position)]

    def add_node(self, node: Node):
        self.nodes.add(node)

    def add_edge(self, from_node: Node, to_node: Node, distance: int, on_road: bool=True):
        if to_node.name not in self.edges[from_node.name]:
            self.edges[from_node.name].append(to_node.name)
        if from_node.name not in self.edges[to_node.name]:
            self.edges[to_node.name].append(from_node.name)
        self.edge_properties[(from_node.name, to_node.name)] = (distance, on_road)

    @staticmethod
    def from_pickle(pickle_file):
        with open(pickle_file, 'rb') as f:
            return pickle.load(f)

    def _dijkstra(self, initial):
        visited = {initial: 0}
        path = {}

        nodes = self.node_names()

        while nodes:
            min_node = None
            for node in nodes:
                if node in visited:
                    if min_node is None:
                        min_node = node
                    elif visited[node] < visited[min_node]:
                        min_node = node
            if min_node is None:
                break

            nodes.remove(min_node)
            current_weight = visited[min_node]

            for edge in self.edges[min_node]:
                try:
                    weight = current_weight + self.edge_properties[(min_node, edge)][0]
                except KeyError:
                    continue
                if edge not in visited or weight < visited[edge]:
                    visited[edge] = weight
                    path[edge] = min_node

        return visited, path

    def shortest_path(self, origin, destination) -> Tuple[int, List[str]]:
        visited, paths = self._dijkstra(origin)
        full_path = deque()
        _destination = paths[destination]

        while _destination != origin:
            full_path.appendleft(_destination)
            _destination = paths[_destination]

        full_path.appendleft(origin)
        full_path.append(destination)

        return visited[destination], list(full_path)

    def travel(self, vehicle_group, from_node: Node, to_node: Node, speed=32):
        distance, path = self.shortest_path(from_node.name, to_node.name)
        last = path[0]
        for p in path[1:]:
            current_node = self.node(p)
            if not self.edge_properties[(last, p)][1]:
                vehicle_group.add_waypoint(current_node.position + mapping.Point(1, 0),
                                           speed=speed,
                                           move_formation=point.PointAction.OffRoad)
                vehicle_group.add_waypoint(current_node.position,
                                           speed=speed,
                                           move_formation=point.PointAction.OnRoad)
            else:
                vehicle_group.add_waypoint(current_node.position,
                                           speed=speed,
                                           move_formation=point.PointAction.OnRoad)
            last = p
        return distance, path

    def store_pickle(self, file_name):
        with open(file_name, 'wb') as file:
            pickle.dump(self, file)

    def __str__(self):
        s = "digraph city_graph {\n"
        for x in self.nodes:
            s += '  "' + x.name + '" -> ' + ",".join(['"' + y + '"' for y in self.edges[x.name]]) + "\n"
        s += "}\n"
        return s


class Terrain:
    bounds = None  # type: mapping.Rectangle
    map_view_default = None  # type: MapView
    city_graph = Graph()  # type: Graph

    def __init__(self, name: str):
        self.name = name
        self.center = {"lat": 0, "long": 0}  # WGS84 decimal
        self.bullseye_blue = {"x": 0, "y": 0}
        self.bullseye_red = {"x": 0, "y": 0}
        self.airports = {}  # type: Dict[str,Airport]

    def airport_by_id(self, id: int) -> Airport:
        for x in self.airports:
            if self.airports[x].id == id:
                return self.airports[x]
        return None

    def airport_list(self) -> List[Airport]:
        for x in self.airports:
            yield self.airports[x]


class Warehouses:
    def __init__(self, terrain: Terrain):
        self.terrain = terrain
        self.warehouses = {}

    def load_dict(self, data):
        for x in data.get("airports", {}):
            self.terrain.airport_by_id(x).load_from_dict(data["airports"][x])

    def __str__(self):
        airports = self.terrain.airports
        d = {
            "warehouses": self.warehouses,
            "airports": {airports[x].id: airports[x].dict() for x in airports}
        }
        return lua.dumps(d, "warehouses", 1)



from enum import Enum


class ForcedOptions:

    class Views(Enum):
        OnlyMap = "optview_onlymap"
        MyAircraft = "optview_myaircraft"
        Allies = "optview_allies"
        OnlyAllies = "optview_onlyallies"
        All = "optview_all"

    class CivilTraffic(Enum):
        Off = ""
        Low = "low"
        Medium = "medium"
        High = "high"

    class GEffect(Enum):
        None_ = ""
        Game = "game"
        Realistic = "realistic"

    def __init__(self):
        self.fuel = None
        self.easy_radar = None
        self.mini_hud = None
        self.accidental_failures = None
        self.options_view = None  # type: ForcedOptions.Views
        self.permit_crash = None
        self.immportal = None
        self.easy_communication = None
        self.cockpit_visual_recon_mode = None
        self.easy_flight = None
        self.radio = None
        self.tips = None
        self.geffect = None
        self.external_views = None
        self.birds = None
        self.civil_traffic = None  # type: ForcedOptions.CivilTraffic
        self.weapons = None
        self.padlock = None
        self.scenes = None
        self.labels = None

    def load_from_dict(self, d):
        self.fuel = d.get("fuel")
        self.easy_radar = d.get("easyRadar")
        self.mini_hud = d.get("miniHUD")
        self.accidental_failures = d.get("accidental_failures")
        if d.get("optionsView") is not None:
            self.options_view = ForcedOptions.Views(d["optionsView"])
        self.permit_crash = d.get("permitCrash")
        self.immportal = d.get("immortal")
        self.easy_communication = d.get("easyCommunication")
        self.cockpit_visual_recon_mode = d.get("cockpitVisualRM")
        self.easy_flight = d.get("easyFlight")
        self.radio = d.get("radio")
        self.tips = d.get("tips")
        if d.get("geffect") is not None:
            self.geffect = ForcedOptions.GEffect(d["geffect"])
        self.external_views = d.get("externalViews")
        self.birds = d.get("birds")
        if d.get("civTraffic") is not None:
            self.civil_traffic = ForcedOptions.CivilTraffic(d["civTraffic"])
        self.weapons = d.get("weapons")
        self.padlock = d.get("padlock")
        self.labels = d.get("labels")
        self.scenes = d.get("scenes")

    def dict(self):
        d = {}
        if self.fuel is not None:
            d["fuel"] = self.fuel
        if self.easy_radar is not None:
            d["easyRadar"] = self.easy_radar
        if self.mini_hud is not None:
            d["miniHUD"] = self.mini_hud
        if self.accidental_failures is not None:
            d["accidental_failures"] = self.accidental_failures
        if self.options_view is not None:
            d["optionsView"] = self.options_view.value
        if self.permit_crash is not None:
            d["permitCrash"] = self.permit_crash
        if self.immportal is not None:
            d["immortal"] = self.immportal
        if self.easy_communication is not None:
            d["easyCommunication"] = self.easy_communication
        if self.cockpit_visual_recon_mode is not None:
            d["cockpitVisualRM"] = self.cockpit_visual_recon_mode
        if self.easy_flight is not None:
            d["easyFlight"] = self.easy_flight
        if self.radio is not None:
            d["radio"] = self.radio
        if self.tips is not None:
            d["tips"] = self.tips
        if self.geffect is not None:
            d["geffect"] = self.geffect.value
        if self.external_views is not None:
            d["externalViews"] = self.external_views
        if self.birds is not None:
            d["birds"] = self.birds
        if self.civil_traffic is not None:
            d["civTraffic"] = self.civil_traffic.value
        if self.weapons is not None:
            d["weapons"] = self.weapons
        if self.padlock is not None:
            d["padlock"] = self.padlock
        if self.labels is not None:
            d["labels"] = self.labels
        if self.scenes is not None:
            d["scenes"] = self.scenes
        return d



"""This module holds all Tasks that are possible to specifiy in dcs.

There are 2 type of tasks, a MainTask and a Task action.
 * MainTasks are the flight groups main objective like :py:class:`CAS`, :py:class:`CAP`, :py:class:`SEAD`, ...
 * Task actions on the otherhand are specific tasks within a MainTask, these
   cover things like :py:class:`AttackGroup`, :py:class:`Orbit`, :py:class:`Follow`, :py:class:`Escort`, ...

Also options and commands are task actions.
"""
from typing import List, Dict, Optional
from enum import Enum
from .mapping import Point


def _create_from_dict(d):
    _id = d["id"]
    if _id == "WrappedAction":
        actionid = d["params"]["action"]["id"]
        if actionid == "Option":
            t = options[d["params"]["action"]["params"]["name"]].create_from_dict(d)
        else:
            t = wrappedactions[actionid].create_from_dict(d)
    elif _id == "EngageTargets":
        t = engagetargets_tasks[d["key"]].create_from_dict(d)
    else:
        t = tasks_map[_id].create_from_dict(d)

    t.auto = d["auto"]
    t.enabled = d["enabled"]
    t.number = d["number"]
    t.params = d["params"]
    return t


class Modulation(Enum):
    """Enum for VHF frequency bands"""
    AM = 0
    """AM frequency band"""
    FM = 1
    """FM frequency band"""


class Task:
    """Base class for task actions."""

    def __init__(self, _id):
        self.id = _id
        self.params = {}
        self.auto = False
        self.number = 1
        self.enabled = True

    @classmethod
    def create_from_dict(cls, d):
        t = cls()
        t.params = d["params"]
        return t

    def dict(self):
        return {
            "id": self.id,
            "auto": self.auto,
            "enabled": self.enabled,
            "params": self.params,
            "number": self.number
        }


class ControlledTask(Task):
    """A ControlledTask is a task action with start and stop conditions.

    ControlledTask is a wrapper around a normal task action that has special
    methods to add start/stop conditions.

    :param task: to wrap
    """
    Id = "ControlledTask"

    def __init__(self, task: Task=None):
        super(ControlledTask, self).__init__(self.Id)
        if task:
            self.params["task"] = task.dict()

    def start_after_time(self, time: int):
        """Start the wrapped task after time seconds.

        :param time: start after x seconds.
        """
        self.params.setdefault("condition", {})
        self.params["condition"]["time"] = time

    def start_if_user_flag(self, user_flag, value: bool):
        """Start the wrapped task if user_flag has value.

        :param user_flag: id of the userflag
        :param value: bool value of the flag
        """
        self.params.setdefault("condition", {})
        self.params["condition"]["userFlag"] = str(user_flag)
        self.params["condition"]["userFlagValue"] = value

    def start_probability(self, probability: int):
        """Chance that the wrapped task will actually start.

        :param probability: start chance in %
        """
        self.params.setdefault("condition", {})
        self.params["condition"]["probability"] = probability

    def start_if_lua_predicate(self, lua_predicate: str):
        """Start wrapped task if lua condition is true.

        :param lua_predicate: lua condition as string
        """
        self.params.setdefault("condition", {})
        self.params["condition"]["condition"] = lua_predicate

    def stop_after_time(self, time: int):
        """Stop the wrapped task after time seconds.

        :param time: start after x seconds.
        """
        self.params.setdefault("stopCondition", {})
        self.params["stopCondition"]["time"] = time

    def stop_if_user_flag(self, user_flag, value: bool):
        """Stop the wrapped task if user_flag has value.

        :param user_flag: id of the userflag
        :param value: bool value of the flag
        """
        self.params.setdefault("stopCondition", {})
        self.params["stopCondition"]["userFlag"] = str(user_flag)
        self.params["stopCondition"]["userFlagValue"] = value

    def stop_if_lua_predicate(self, lua_predicate: str):
        """Stop wrapped task if lua condition is true.

        :param lua_predicate: lua condition as string
        """
        self.params.setdefault("stopCondition", {})
        self.params["stopCondition"]["condition"] = lua_predicate

    def stop_after_duration(self, duration: int):
        """Stop task after duration seconds.

        :param duration: in seconds
        """
        self.params.setdefault("stopCondition", {})
        self.params["stopCondition"]["duration"] = duration


class WeaponType(Enum):
    Auto = 1073741822


class TargetType(type):
    id = None

    def __str__(self):
        return '"{id}"'.format(id=self.id)


class Targets(metaclass=TargetType):
    class All(metaclass=TargetType):
        id = "All"
        class Air(metaclass=TargetType):
            id = "Air"
            class Planes(metaclass=TargetType):
                id = "Planes"
                class Fighters(metaclass=TargetType):
                    id = "Fighters"
                class Bombers(metaclass=TargetType):
                    id = "Bombers"
            class Helicopters(metaclass=TargetType):
                id = "Helicopters"

        class GroundUnits(metaclass=TargetType):
            id = "Ground Units"
            class Infantry(metaclass=TargetType):
                id = "Infantry"
            class Fortifications(metaclass=TargetType):
                id = "Fortifications"
            class GroundVehicles(metaclass=TargetType):
                id = "Ground vehicles"

                class ArmoredVehicles(metaclass=TargetType):
                    id = "Armored vehicles"
                    class Tanks(metaclass=TargetType):
                        id = "Tanks"
                    class IFV(metaclass=TargetType):
                        id = "IFV"
                    class APC(metaclass=TargetType):
                        id = "APC"
                class Artillery(metaclass=TargetType):
                    id = "Artillery"
                class UnarmedVehicles(metaclass=TargetType):
                    id = "Unarmed vehicles"
            class AirDefence(metaclass=TargetType):
                id = "Air Defence"
                class AAA(metaclass=TargetType):
                    id = "AAA"
                    class SAMRelated(metaclass=TargetType):
                        id = "SAM related"
                        class SRSAM(metaclass=TargetType):
                            id = "SR SAM"
                        class MRSAM(metaclass=TargetType):
                            id = "MR SAM"
                        class LRSAM(metaclass=TargetType):
                            id = "LR SAM"

        class Naval(metaclass=TargetType):
            id = "Naval"
            class Ships(metaclass=TargetType):
                id = "Ships"
                class ArmedShips(metaclass=TargetType):
                    id = "Armed ships"
                    class HeavyArmedShips(metaclass=TargetType):
                        id = "Heavy armed ships"
                        class AircraftCarriers(metaclass=TargetType):
                            id = "Aircraft Carriers"
                        class Cruisers(metaclass=TargetType):
                            id = "Cruisers"
                        class Destroyers(metaclass=TargetType):
                            id = "Destroyers"
                        class Frigates(metaclass=TargetType):
                            id = "Frigates"
                        class Corvettes(metaclass=TargetType):
                            id = "Corvettes"
                    class LightArmedShips(metaclass=TargetType):
                        id = "Light armed ships"
                class UnarmedShips(metaclass=TargetType):
                    id = "Unarmed ships"
            class Submarines(metaclass=TargetType):
                id = "Submarines"


class NoTask(Task):
    Id = "NoTask"

    def __init__(self):
        super(NoTask, self).__init__(self.Id)


class AttackGroup(Task):
    """Attack group task action

    :param group_id: group id to attack
    :param weapon_type: :py:class:`WeaponType` to use for the attack
    """
    Id = "AttackGroup"

    def __init__(self, group_id=0, weapon_type: WeaponType=WeaponType.Auto):
        super(AttackGroup, self).__init__(AttackGroup.Id)
        self.params = {
            "groupId": group_id,
            "weaponType": weapon_type.value
        }


class AttackUnit(Task):
    """Attack unit task action

    :param unit_id: unit id to attack
    :param attack_limit: how much attack runs
    :param weapon_type: :py:class:`WeaponType` to use for the attack
    :param group_attack: indicate if the unit must be attacked by all aircrafts in the group.
    """
    Id = "AttackUnit"

    def __init__(self, unit_id=0, attack_limit: Optional[int]=None,
                 weapon_type: WeaponType=WeaponType.Auto, group_attack=False):
        super(AttackUnit, self).__init__(self.Id)
        self.params = {
            "groupId": unit_id,
            "weaponType": weapon_type.value,
            "groupAttack": group_attack,
            "attackQty": attack_limit,
            "attackQtyLimit": attack_limit is not None
        }


class AttackMapObject(Task):
    Id = "AttackMapObject"

    def __init__(self, position: Point=Point(0, 0), attack_limit: Optional[int]=None,
                 weapon_type: WeaponType=WeaponType.Auto, group_attack=False):
        super(AttackMapObject, self).__init__(self.Id)
        self.params = {
            "x": position.x,
            "y": position.y,
            "weaponType": weapon_type.value,
            "groupAttack": group_attack,
            "attackQty": attack_limit,
            "attackQtyLimit": attack_limit is not None
        }


class AntishipStrikeTaskAction(Task):
    Id = "EngageTargets"
    Key = "AntiShip"

    def __init__(self):
        super(AntishipStrikeTaskAction, self).__init__(AntishipStrikeTaskAction.Id)
        self.params = {
            "targetTypes": {
                1: Targets.All.Naval.Ships
            },
            "priority": 0
        }

    def dict(self):
        d = super(AntishipStrikeTaskAction, self).dict()
        d["key"] = AntishipStrikeTaskAction.Key
        return d


class CASTaskAction(Task):
    Id = "EngageTargets"
    Key = "CAS"

    def __init__(self):
        super(CASTaskAction, self).__init__(CASTaskAction.Id)
        self.params = {
            "targetTypes": {
                1: Targets.All.Air.Helicopters,
                2: Targets.All.GroundUnits,
                3: Targets.All.Naval.Ships.ArmedShips.LightArmedShips
            },
            "priority": 0
        }

    def dict(self):
        d = super(CASTaskAction, self).dict()
        d["key"] = CASTaskAction.Key
        return d


class SEADTaskAction(Task):
    Id = "EngageTargets"
    Key = "SEAD"

    def __init__(self):
        super(SEADTaskAction, self).__init__(SEADTaskAction.Id)
        self.params = {
            "targetTypes": {1: Targets.All.GroundUnits.AirDefence},
            "priority": 0
        }

    def dict(self):
        d = super(SEADTaskAction, self).dict()
        d["key"] = SEADTaskAction.Key
        return d


class CAPTaskAction(Task):
    Id = "EngageTargets"
    Key = "CAP"

    def __init__(self):
        super(CAPTaskAction, self).__init__(CAPTaskAction.Id)
        self.params = {
            "targetTypes": {1: Targets.All.Air},
            "priority": 0
        }

    def dict(self):
        d = super(CAPTaskAction, self).dict()
        d["key"] = CAPTaskAction.Key
        return d


class FighterSweepTaskAction(Task):
    Id = "EngageTargets"
    Key = "FighterSweep"

    def __init__(self):
        super(FighterSweepTaskAction, self).__init__(FighterSweepTaskAction.Id)
        self.params = {
            "targetTypes": {1: Targets.All.Air.Planes},
            "priority": 0
        }

    def dict(self):
        d = super(FighterSweepTaskAction, self).dict()
        d["key"] = FighterSweepTaskAction.Key
        return d

engagetargets_tasks = {
    AntishipStrikeTaskAction.Key: AntishipStrikeTaskAction,
    CASTaskAction.Key: CASTaskAction,
    CAPTaskAction.Key: CAPTaskAction,
    SEADTaskAction.Key: SEADTaskAction,
    FighterSweepTaskAction.Key: FighterSweepTaskAction
}


class EscortTaskAction(Task):
    Id = "Escort"

    def __init__(self,
                 group_id=None,
                 engagement_max_dist=60000,
                 lastwpt=None,
                 targets: List[str]=None,
                 position: Dict[str, float]=None):
        super(EscortTaskAction, self).__init__(EscortTaskAction.Id)
        if targets is None:
            targets = [Targets.All.Air.Planes]
        if position is None:
            position = {"x": -200, "y": -100, "z": -500}
        self.params = {
            "lastWptIndexFlagChangedManually": False,
            "lastWptIndexFlag": False,
            "engagementDistMax": engagement_max_dist,
            "targetTypes": {i: targets[i-1] for i in range(1, len(targets)+1)},
            "pos": position
        }
        if group_id:
            self.params["groupId"] = group_id
        if lastwpt:
            self.params["lastWptIndexFlagChangedManually"] = True
            self.params["lastWptIndexFlag"] = True
            self.params["lastWptIndex"] = lastwpt


class Expend(Enum):
    Auto = "Auto"
    One = "One"
    Two = "Two"
    Four = "Four"
    Quarter = "Quarter"
    Half = "Half"


class Bombing(Task):
    Id = "Bombing"

    def __init__(self, position: Point=Point(0, 0), weapon_type: WeaponType=WeaponType.Auto,
                 expend: Expend=Expend.Auto, attack_qty=1, group_attack=False,
                 direction: Optional[int]=None, altitude: Optional[int]=None):
        super(Bombing, self).__init__(Bombing.Id)
        self.params = {
            "directionEnabled": direction is not None,
            "direction": direction if direction is not None else 0,
            "attackQtyLimit": attack_qty > 1,
            "attackQty": attack_qty if attack_qty > 1 else 1,
            "expend": expend.value,
            "x": position.x,
            "y": position.y,
            "groupAttack": group_attack,
            "altitude": altitude if altitude is not None else 0,
            "altitudeEnabled": altitude is not None,
            "weaponType": weapon_type.value
        }


class BombingRunway(Task):
    Id = "BombingRunway"

    def __init__(self, airport_id: int=0, weapon_type: WeaponType=WeaponType.Auto,
                 expend: Expend=Expend.Auto, attack_qty=1, group_attack=False,
                 direction: Optional[int]=None, altitude: Optional[int]=None):
        super(BombingRunway, self).__init__(BombingRunway.Id)
        self.params = {
            "directionEnabled": direction is not None,
            "direction": direction if direction is not None else 0,
            "attackQtyLimit": attack_qty > 1,
            "attackQty": attack_qty if attack_qty > 1 else 1,
            "expend": expend.value,
            "runwayId": airport_id,
            "groupAttack": group_attack,
            "altitude": altitude if altitude is not None else 0,
            "altitudeEnabled": altitude is not None,
            "weaponType": weapon_type.value
        }


class EngageTargets(Task):
    Id = "EngageTargets"

    def __init__(self, max_distance=None, targets: List[str]=None):
        super(EngageTargets, self).__init__(EngageTargets.Id)
        if targets is None:
            targets = [Targets.All]
        self.params = {
            "targetTypes": {i: targets[i-1] for i in range(1, len(targets)+1)},
            "maxDistEnabled": True if max_distance else False,
            "maxDist": max_distance,
            "priority": 0
        }


class EngageTargetsInZone(Task):
    Id = "EngageTargetsInZone"

    def __init__(self, position: Point=Point(0, 0), radius=5000, targets: List[str]=None):
        super(EngageTargetsInZone, self).__init__(EngageTargetsInZone.Id)
        if targets is None:
            targets = [Targets.All]
        self.params = {
            "targetTypes": {i: targets[i-1] for i in range(1, len(targets)+1)},
            "priority": 0,
            "x": position.x,
            "y": position.y,
            "zoneRadius": radius
        }


class EngageGroup(Task):
    Id = "EngageGroup"

    def __init__(self, group_id=0, visible=False):
        super(EngageGroup, self).__init__(EngageGroup.Id)
        self.auto = False
        self.params = {
            "visible": visible,
            "groupId": group_id,
            "priority": 1,
            "weaponType": 1073741822
        }


class EngageUnit(Task):
    Id = "EngageUnit"

    def __init__(self, unit_id=0, visible=False):
        super(EngageUnit, self).__init__(EngageUnit.Id)
        self.auto = False
        self.params = {
            "visible": visible,
            "groupAttack": False,
            "unitId": unit_id,
            "priority": 1,
            "weaponType": 1073741822,
            "directionEnabled": False,
            "direction": 0,
            "altitudeEnabled": False,
            "attackQtyLimit": False,
            "attackQty": 0
        }


class FireAtPoint(Task):
    Id = "FireAtPoint"

    def __init__(self, position: Point=Point(0, 0), rounds=None, radius=0):
        super(FireAtPoint, self).__init__(FireAtPoint.Id)
        self.auto = False
        self.params = {
            "y": position.y,
            "x": position.x,
            "expendQty": rounds if rounds else 1,
            "expendQtyEnabled": rounds is not None,
            "templateId": "",
            "weaponType": WeaponType.Auto.value,
            "zoneRadius": radius
        }


class Hold(Task):
    """Unit will hold current position"""
    Id = "Hold"

    def __init__(self):
        super(Hold, self).__init__(self.Id)

        self.params = {
            "templateId": ""
        }


class AWACSTaskAction(Task):
    Id = "AWACS"

    def __init__(self):
        super(AWACSTaskAction, self).__init__(AWACSTaskAction.Id)


class RefuelingTaskAction(Task):
    """Assigns the aircraft group to refuel at the nearest tanker aircraft."""
    Id = "Refueling"

    def __init__(self):
        super(RefuelingTaskAction, self).__init__(RefuelingTaskAction.Id)


class Tanker(Task):
    """Assigns the aircraft to act as an Airborne tanker."""
    Id = "Tanker"

    def __init__(self):
        super(Tanker, self).__init__(Tanker.Id)


class OrbitAction(Task):
    Id = "Orbit"

    class OrbitPattern(Enum):
        RaceTrack = "Race-Track"
        Circle = "Circle"

    def __init__(self, altitude=4000, speed=600, pattern: OrbitPattern=OrbitPattern.RaceTrack):
        super(OrbitAction, self).__init__(OrbitAction.Id)
        self.params = {
            "altitude": altitude,
            "pattern": pattern.value,
            "speed": speed / 3.6,
            "speedEdited": True
        }


class Follow(Task):
    Id = "Follow"

    def __init__(self, groupid=None, position: Point=Point(-200, 0), altitude_difference=-200, last_wpt=None):
        super(Follow, self).__init__(self.Id)

        self.params = {
            "pos": {"x": position.x, "z": position.y, "y": altitude_difference}
        }
        if groupid is not None:
            self.params["groupId"] = groupid
        if last_wpt:
            self.params["lastWptIndexFlag"] = True
            self.params["lastWptIndex"] = last_wpt
            self.params["lastWptIndexFlagChangedManually"] = False


class Aerobatics(Task):
    Id = "Aerobatics"

    def __init__(self):
        super(Aerobatics, self).__init__(self.Id)

        self.params = {
            "maneuversSequency": [],  # TODO
            "maneuversParams": {}
        }


class Designation(Enum):
    No = "No"
    Auto = "Auto"
    WP = "WP"
    IRPointer = "IR-Pointer"
    Laser = "Laser"


class FAC(Task):
    Id = "FAC"

    def __init__(self, callsign: int=1, designation: Designation=Designation.Auto,
                 frequency: int=30, modulation: Modulation=Modulation.FM, number: int=1):
        super(FAC, self).__init__(self.Id)

        self.params = {
            "designation": designation.value,
            "frequency": frequency * 1000000,
            "modulation": modulation.value,
            "datalink": True,
            "callname": callsign,
            "number": number
        }


class FACEngageGroup(Task):
    Id = "FAC_EngageGroup"

    def __init__(self, group_id: int=0, visible=False, weapon_type: WeaponType=WeaponType.Auto, priority: int=0,
                 callsign: int=1, designation: Designation=Designation.Auto,
                 frequency: int=30, modulation: Modulation=Modulation.FM, datalink=True, number: int=1):
        super(FACEngageGroup, self).__init__(self.Id)

        self.params = {
            "groupId": group_id,
            "visible": visible,
            "weaponType": weapon_type.value,
            "designation": designation.value,
            "frequency": frequency * 1000000,
            "modulation": modulation.value,
            "datalink": datalink,
            "callname": callsign,
            "number": number,
            "priority": priority
        }


class Land(Task):
    """Helicopter landing task.

    If added to a helicopter group, the group will land at the given coordinates for the given duration.

    :param position: :py:class:`dcs.mapping.Point` where to land
    :param duration: how long the helicopter should stay on ground in seconds.
    """
    Id = "Land"

    def __init__(self, position: Point=Point(0, 0), duration: int=None):
        super(Land, self).__init__(self.Id)

        self.params = {
            "x": position.x,
            "y": position.y,
            "duration": duration if duration else 300,
            "durationFlag": duration is not None
        }


class Embarking(Task):
    """Pickup task for helicopters.

    Helicopter group will land at the given coordinates and will pickup the given groups.

    :param position: :py:class:`dcs.mapping.Point` where to land and pickup
    :param groupids: list of groups to pickup
    :param distribution: dictionary with heli unit to groups to pickup mapping {heliunit: [grp1, grp2], ..}
    :param duration: how long the helicopter should stay on ground  and wait in seconds.
    """
    Id = "Embarking"

    def __init__(self, position: Point=Point(0, 0), groupids: List[int]=None,
                 distribution: Dict[int, List[int]]=None, duration: int=None):
        super(Embarking, self).__init__(self.Id)

        groupids = [] if groupids is None else groupids
        distribution = {} if distribution is None else distribution

        self.params = {
            "x": position.x,
            "y": position.y,
            "duration": duration if duration else 300,
            "durationFlag": duration is not None,
            "groupsForEmbarking": {x: x for x in groupids},
            "distributionFlag": len(distribution) > 0,
            "distribution": {x: {y: y for y in distribution[x]} for x in distribution}
        }


class EmbarkToTransport(Task):
    """Task for ground units that will get picked up by a helicopter

    :param position: :py:class:`dcs.mapping.Point` where to wait to get picked up.
    :param zone_radius: radius around the point where the group will embark.
    :param concrete_unitid: if specified the group will embark to exaclty this unit.
    """
    Id = "EmbarkToTransport"

    def __init__(self, position: Point=Point(0, 0), zone_radius=200, concrete_unitid=None):
        super(EmbarkToTransport, self).__init__(self.Id)

        self.params = {
            "x": position.x,
            "y": position.y,
            "zoneRadius": zone_radius,
            "concretteUnitChecked": concrete_unitid is not None,
            "selectedUnit": concrete_unitid if concrete_unitid else -1
        }


class DisembarkFromTransport(Task):
    """Task for ground units that will disembark a transport helicopter

    :param position: :py:class:`dcs.mapping.Point` where the group will disembark
    :param zone_radius: radius around the point where the group will disembark.
    """
    Id = "DisembarkFromTransport"

    def __init__(self, position: Point=Point(0, 0), zone_radius=200):
        super(DisembarkFromTransport, self).__init__(self.Id)

        self.params = {
            "x": position.x,
            "y": position.y,
            "zoneRadius": zone_radius
        }


class CargoTransportation(Task):
    """Task for Cargo transportation.

    :param groupid: cargo group id
    :param zoneid: zone id to transport to??
    """
    Id = "CargoTransportation"

    def __init__(self, groupid=None, zoneid=None):
        super(CargoTransportation, self).__init__(self.Id)

        self.params = {}
        if groupid:
            self.params["groupId"] = groupid
        if zoneid:
            self.params["zoneId"] = zoneid


tasks_map = {
    ControlledTask.Id: ControlledTask,
    EscortTaskAction.Id: EscortTaskAction,
    AttackGroup.Id: AttackGroup,
    Bombing.Id: Bombing,
    BombingRunway.Id: BombingRunway,
    EngageTargetsInZone.Id: EngageTargetsInZone,
    EngageGroup.Id: EngageGroup,
    EngageUnit.Id: EngageUnit,
    AWACSTaskAction.Id: AWACSTaskAction,
    RefuelingTaskAction.Id: RefuelingTaskAction,
    Tanker.Id: Tanker,
    OrbitAction.Id: OrbitAction,
    Follow.Id: Follow,
    Aerobatics.Id: Aerobatics,
    FAC.Id: FAC,
    FACEngageGroup.Id: FACEngageGroup,
    Hold.Id: Hold,
    Land.Id: Land,
    Embarking.Id: Embarking,
    EmbarkToTransport.Id: EmbarkToTransport,
    DisembarkFromTransport.Id: DisembarkFromTransport,
    CargoTransportation.Id: CargoTransportation
}


class WrappedAction(Task):
    Id = "WrappedAction"

    def __init__(self):
        super(WrappedAction, self).__init__(WrappedAction.Id)


class EPLRS(WrappedAction):
    Key = "EPLRS"

    def __init__(self, group_id=1):
        super(EPLRS, self).__init__()
        self.params = {
            "action": {"id": EPLRS.Key, "params": {"value": True, "groupId": group_id}}
        }

    @property
    def eplrs(self):
        return self.params["action"]["params"]["groupId"]


class ActivateBeaconCommand(WrappedAction):
    Key = "ActivateBeacon"

    @staticmethod
    def calc_tacan_frequency(mode_channel, channel, aa=True):
        if not aa and mode_channel == "X":
            if channel < 64:
                freq = 962 + channel - 1
            else:
                freq = 1151 + channel - 64
        else:
            if channel < 64:
                freq = 1088 + channel - 1
            else:
                freq = 1025 + channel - 64

        return freq * 1000000

    def __init__(self, channel=10, modechannel="X", callsign="TKR", bearing=True):
        super(ActivateBeaconCommand, self).__init__()
        self.params = {
            "action": {
                "id": ActivateBeaconCommand.Key,
                "params": {
                    "type": 4,
                    "frequency": self.calc_tacan_frequency(modechannel, channel),
                    "callsign": callsign,
                    "channel": channel,
                    "modeChannel": modechannel,
                    "bearing": bearing,
                    "system": 4
                }
            }
        }


class RunScript(WrappedAction):
    """Runs a given script string

    :param script: to be executed
    """
    Key = "Script"

    def __init__(self, script: str=""):
        super(RunScript, self).__init__()
        self.params = {
            "action": {
                "id": self.Key,
                "params": {"command": script}
            }
        }


class RunScriptFile(WrappedAction):
    """Runs a script attached to the mission

    :param resourcekey: resource key to the script file, see :py:class:`dcs.mission.MapResource`
    """
    Key = "ScriptFile"

    def __init__(self, resourcekey: str=""):
        super(RunScriptFile, self).__init__()
        self.params = {
            "action": {
                "id": self.Key,
                "params": {"file": resourcekey}
            }
        }


class TransmitMessage(WrappedAction):
    """Transmits a given sound file over the current radio channel.

    :param soundfile_reskey: resource key to the sound file to transmit, see :py:class:`dcs.mission.MapResource`
    :param subtitle_resstring: string resource key to subtitle displayed,
                               see :py:attr:`dcs.mission.Mission.translation`
    :param loop: True or False if the sound file should be looped.
    :param subtitle_duration: how long the subtitle should be displayed in seconds.
    """
    Key = "TransmitMessage"

    def __init__(self, soundfile_reskey: Optional[str]=None, subtitle_resstring: Optional[str]=None,
                 loop=False, subtitle_duration=5):
        super(TransmitMessage, self).__init__()

        self.params = {
            "action": {
                "id": self.Key,
                "params": {
                    "file": soundfile_reskey,
                    "loop": loop,
                    "duration": subtitle_duration
                }
            }
        }
        if subtitle_resstring:
            self.params["action"]["params"]["subtitle"] = subtitle_resstring


class StopTransmission(WrappedAction):
    """Stops any :py:class:`dcs.task.TransmitMessage` task currently ongoing.
    """
    Key = "StopTransmission"

    def __init__(self):
        super(StopTransmission, self).__init__()

        self.params = {
            "action": {
                "id": self.Key,
                "params": {}
            }
        }


class SetFrequencyCommand(WrappedAction):
    """Set the groups radio frequency.

    :param frequency: frequency band in mhz.
    :param modulation: AM or FM, see :py:class:`dcs.task.Modulation`
    """
    Key = "SetFrequency"

    def __init__(self, frequency=133, modulation: Modulation=Modulation.AM):
        super(SetFrequencyCommand, self).__init__()
        self.params = {
            "action": {
                "id": SetFrequencyCommand.Key,
                "params": {"modulation": modulation.value, "frequency": frequency * 1000000}
            }
        }


class SwitchWaypoint(WrappedAction):
    """Switch to a different waypoint.

    :param from_waypoint: from which waypoint to switch.??
    :param to_waypoint: new current waypoint
    """
    Key = "SwitchWaypoint"

    def __init__(self, from_waypoint=1, to_waypoint=2):
        super(SwitchWaypoint, self).__init__()
        self.params = {
            "action": {
                "id": self.Key,
                "params": {
                    "goToWaypointIndex": to_waypoint,
                    "fromWaypointIndex": from_waypoint
                }
            }
        }


class SetInvisibleCommand(WrappedAction):
    Key = "SetInvisible"

    def __init__(self, value=True):
        super(SetInvisibleCommand, self).__init__()
        self.params = {
            "action": {
                "id": SetInvisibleCommand.Key,
                "params": {"value": value}
            }
        }


class SetImmortalCommand(WrappedAction):
    Key = "SetImmortal"

    def __init__(self, value=True):
        super(SetImmortalCommand, self).__init__()
        self.params = {
            "action": {
                "id": SetImmortalCommand.Key,
                "params": {"value": value}
            }
        }

wrappedactions = {
    EPLRS.Key: EPLRS,
    ActivateBeaconCommand.Key: ActivateBeaconCommand,
    SetFrequencyCommand.Key: SetFrequencyCommand,
    SetInvisibleCommand.Key: SetInvisibleCommand,
    SetImmortalCommand.Key: SetImmortalCommand,
    RunScript.Key: RunScript,
    RunScriptFile.Key: RunScriptFile,
    TransmitMessage.Key: TransmitMessage,
    StopTransmission.Key: StopTransmission,
    SwitchWaypoint.Key: SwitchWaypoint
}


class MainTask:
    name = None
    sub_tasks = []
    perform_task = []  # type: List[Task]
    map = {}  # type: Dict[str, MainTask]


class Nothing(MainTask):
    id = 15
    name = "Nothing"
    sub_tasks = [OrbitAction, Follow, Aerobatics]


class AFAC(MainTask):
    id = 16
    name = "AFAC"
    sub_tasks = [OrbitAction, Follow, AttackGroup, AttackUnit, Bombing, AttackMapObject]


class AWACS(MainTask):
    id = 14
    name = "AWACS"
    sub_tasks = [OrbitAction, Follow, RefuelingTaskAction]
    perform_task = [AWACSTaskAction]


class AntishipStrike(MainTask):
    id = 30
    name = "Antiship Strike"
    sub_tasks = [OrbitAction, Follow, AttackGroup, AttackUnit]
    perform_task = [AntishipStrikeTaskAction]


class CAS(MainTask):
    id = 31
    name = "CAS"
    sub_tasks = [OrbitAction, Follow, AttackGroup, AttackUnit, Aerobatics, RefuelingTaskAction]
    perform_task = [CASTaskAction]

    class EnrouteTasks:
        EngageGroup = EngageGroup
        EngageTargetsInZone = EngageTargetsInZone
        EngageUnit = EngageUnit


class CAP(MainTask):
    id = 11
    name = "CAP"
    sub_tasks = [OrbitAction, Follow, Aerobatics]
    perform_task = [CAPTaskAction]

    class EnrouteTasks:
        EngageTargets = EngageTargets
        EngageTargetsInZone = EngageTargetsInZone
        EngageGroup = EngageGroup
        EngageUnit = EngageUnit


class Escort(MainTask):
    id = 18
    name = "Escort"
    sub_tasks = [OrbitAction, Follow, EscortTaskAction]
    perform_task = [EscortTaskAction]


class FighterSweep(MainTask):
    id = 19
    name = "Fighter Sweep"
    sub_tasks = [OrbitAction, Follow, Aerobatics]
    perform_task = [FighterSweepTaskAction]


class GroundAttack(MainTask):
    id = 32
    name = "Ground Attack"
    sub_tasks = [OrbitAction, Follow, Bombing, AttackMapObject, Aerobatics]


class Intercept(MainTask):
    id = 10
    name = "Intercept"
    sub_tasks = [OrbitAction, Follow, AttackGroup, AttackUnit, Aerobatics]


class PinpointStrike(MainTask):
    id = 33
    name = "Pinpoint Strike"
    sub_tasks = [OrbitAction, Follow, Bombing, AttackMapObject]


class Reconnaissance(MainTask):
    id = 17
    name = "Reconnaissance"
    sub_tasks = [OrbitAction, Follow, Aerobatics]
    perform_task = []


class Refueling(MainTask):
    id = 13
    name = "Refueling"
    sub_tasks = [OrbitAction, Follow]
    perform_task = [Tanker]


class RunwayAttack(MainTask):
    id = 34
    name = "Ground Attack"
    sub_tasks = [OrbitAction, Follow, Bombing, BombingRunway, AttackMapObject]
    perform_task = []


class SEAD(MainTask):
    id = 29
    name = "SEAD"
    sub_tasks = [OrbitAction, Follow, AttackGroup, AttackUnit, EscortTaskAction]
    perform_task = [SEADTaskAction]


class Transport(MainTask):
    id = 35
    name = "Transport"
    sub_tasks = [OrbitAction, Follow, Aerobatics]
    perform_task = []


MainTask.map = {
    Nothing.name: Nothing,
    AFAC.name: AFAC,
    AntishipStrike.name: AntishipStrike,
    CAS.name: CAS,
    SEAD.name: SEAD,
    CAP.name: CAP,
    Escort.name: Escort,
    FighterSweep.name: FighterSweep,
    Intercept.name: Intercept,
    PinpointStrike.name: PinpointStrike,
    GroundAttack.name: GroundAttack,
    RunwayAttack.name: RunwayAttack,
    AWACS.name: AWACS,
    Reconnaissance.name: Reconnaissance,
    Refueling.name: Refueling,
    Transport.name: Transport
}

# options


class Option(Task):
    Id = "WrappedAction"
    Key = None

    def __init__(self, value=None):
        super(Option, self).__init__(Option.Id)
        self.params = {
            "action": {"id": "Option", "params": {"name": self.Key}}
        }
        if value:
            self.params["action"]["params"]["value"] = value


class OptROE(Option):
    """
    Sets the groups rules of engagement.
    Ultimately defines whether or not the AI group is allowed to attack.
    This option can override all other tasking. With the 1.5 patch two of the values
    have different names in the mission editor.
    However the behavior is still exactly the same as before,
    its just labeled slightly different. The scripting engine still uses the previous values.
    """
    Key = 0

    class Values:
        WeaponFree = 0
        """
        AI will engage any enemy group it detects.
        Target prioritization is based based on the threat of the target.
        """
        OpenFireWeaponFree = 1
        """AI will engage any enemy group it detects, but will prioritize targets specified in the groups tasking."""
        OpenFire = 2
        """AI will engage only targets specified in its taskings."""
        ReturnFire = 3
        """AI will only engage threats that shoot first."""
        WeaponHold = 4
        """AI will hold fire under all circumstances."""

    def __init__(self, value=Values.WeaponFree):
        super(OptROE, self).__init__(value)


class OptReactOnThreat(Option):
    """
    Defines the allowable action for an airborne group to take in response to a threat.
    This option can have an effect on other tasking.
    """
    Key = 1

    class Values:
        NoReaction = 0
        """No defensive actions will take place to counter threats"""
        PassiveDefense = 1
        """
        AI will use jammers and other countermeasures in an attempt to defeat the threat.
        AI will not attempt a maneuver to defeat a threat.
        """
        EvadeFire = 2
        """
        AI will react by performing defensive maneuvers against incoming threats,
        AI will also use passive defense.
        """
        ByPassAndEscape = 3
        """
        AI will attempt to avoid enemy threat zones all together.
        This includes attempting to fly above or around threats.
        """
        AllowAbortMission = 4
        """If a threat is deemed severe enough the AI will abort its mission and return to base."""

    def __init__(self, value=Values.NoReaction):
        super(OptReactOnThreat, self).__init__(value)


class OptDisparseUnderFire(Option):
    Key = 8

    def __init__(self, value=None):
        super(OptDisparseUnderFire, self).__init__(value)


options = {
    OptDisparseUnderFire.Key: OptDisparseUnderFire,
    OptReactOnThreat.Key: OptReactOnThreat,
    OptROE.Key: OptROE
}



import math
import random
import copy
from typing import List, Tuple, Union


def point_from_heading(_x, _y, heading, distance):
    """Calculates a point from a given point, heading and distance.

    :param _x: source point x
    :param _y: source point y
    :param heading: heading in degrees from source point
    :param distance: distance from source point
    :return: returns a tuple (x, y) of the calculated point
    """
    while heading < 0:
        heading += 360
    heading %= 360
    rad_heading = math.radians(heading)
    x = _x + math.cos(rad_heading) * distance
    y = _y + math.sin(rad_heading) * distance

    return x, y


def distance(x1, y1, x2, y2):
    """Returns the distance between 2 points

    :param x1: x coordinate of point 1
    :param y1: y coordinate of point 1
    :param x2: x coordinate of point 2
    :param y2: y coordinate of point 2
    :return: distance in point units(m)
    """
    return math.hypot(x2 - x1, y2 - y1)


def heading_between_points(x1, y1, x2, y2):
    """Returns the angle between 2 points in degrees.

    :param x1: x coordinate of point 1
    :param y1: y coordinate of point 1
    :param x2: x coordinate of point 2
    :param y2: y coordinate of point 2
    :return: angle in degrees
    """
    def angle_trunc(a):
        while a < 0.0:
            a += math.pi * 2
        return a
    deltax = x2 - x1
    deltay = y2 - y1
    return math.degrees(angle_trunc(math.atan2(deltay, deltax)))


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def point_from_heading(self, heading, distance):
        x, y = point_from_heading(self.x, self.y, heading, distance)
        return Point(x, y)

    def heading_between_point(self, point):
        return heading_between_points(self.x, self.y, point.x, point.y)

    def distance_to_point(self, point):
        return distance(self.x, self.y, point.x, point.y)

    def random_point_within(self, distance, min_distance=0):
        """Returns a random point within the given distance.

        This is a shortcut for Rectangle.from_point().random_point().

        Args:
            distance: max distance for the random point.
            min_distance: minimum distance the random point should have from origin

        Returns:
            Point: a new random point within the given distance from the point.
        """
        return self.point_from_heading(random.randrange(0, 360),
                                       random.random() * (distance - min_distance) + min_distance)

    def __add__(self, other):
        return Point(self.x + other.x, self.y + other.y)

    def __radd__(self, other):
        if isinstance(other, Point):
            return self + other
        return Point(self.x + other, self.y + other)

    def __sub__(self, other):
        if isinstance(other, Point):
            return Point(self.x - other.x, self.y - other.y)
        return Point(self.x - other, self.y - other)

    def __mul__(self, other):
        return Point(self.x * other, self.y * other)

    def __rmul__(self, other):
        return self * other

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "Point({x}, {y})".format(x=self.x, y=self.y)


class Triangle:
    def __init__(self, points: Union[Tuple[Point, Point, Point], List[Point]]):
        if len(points) != 3:
            raise RuntimeError("Triangle needs 3 points.")
        self.points = copy.copy(points)  # type: List[Point]

    def area(self):
        a = (self.points[0].x * self.points[1].y + self.points[1].x * self.points[2].y +
             self.points[2].x * self.points[0].y - self.points[0].y * self.points[1].x -
             self.points[1].y * self.points[2].x - self.points[2].y * self.points[0].x)
        a /= 2
        return a

    def random_point(self) -> Point:
        r = random.random()
        s = random.random()
        if r + s >= 1:
            r = 1 - r
            s = 1 - s

        return self.points[0] + (r * (self.points[1] - self.points[0]) + s * (self.points[2] - self.points[0]))

    def __repr__(self):
        return "Triangle({points})".format(points=", ".join(map(repr, self.points)))


class Rectangle:
    def __init__(self, top, left, bottom, right):
        self.top = top
        self.left = left
        self.bottom = bottom
        self.right = right

    @staticmethod
    def from_point(point: Point, side_length):
        top = point.x + side_length / 2
        left = point.y - side_length / 2
        bottom = point.x - side_length / 2
        right = point.y + side_length / 2
        return Rectangle(top, left, bottom, right)

    def point_in_rect(self, point: Point):
        return self.bottom <= point.x <= self.top and self.left <= point.y <= self.right

    def height(self):
        return self.top - self.bottom

    def width(self):
        return self.right - self.left

    def center(self) -> Point:
        return Point(self.bottom + (self.height() / 2), self.left + (self.width() / 2))

    def resize(self, percentage: float):
        w = self.width()
        h = self.height()
        w *= percentage / 2
        h *= percentage / 2
        c = self.center()
        return Rectangle(c.x + h, c.y - w, c.x - h, c.y + w)

    def random_point(self) -> Point:
        x = self.bottom + random.random() * (self.top - self.bottom)
        y = self.left + random.random() * (self.right - self.left)
        return Point(x, y)

    def random_distant_points(self, distance) -> Tuple[Point, Point]:
        # determine vertical/horizontal
        if self.width() > self.height():
            axis_y = self.width()
            axis_x = self.height()
            sy = self.left
            sx = self.bottom
            hdg_start = 60
        else:
            axis_y = self.height()
            axis_x = self.width()
            sy = self.bottom
            sx = self.left
            hdg_start = 330

        d = distance if distance < axis_y else axis_y * 0.2
        sy += random.random() * (axis_y - d)
        sx += random.random() * axis_x
        p1 = Point(sx, sy)
        while True:
            hdg = random.random() * 60
            p2 = p1.point_from_heading(hdg_start + hdg, d)
            if self.point_in_rect(p2):
                return p1, p2

    def __eq__(self, other):
        return self.top == other.top and self.bottom == other.bottom \
            and self.left == other.left and self.right == other.right

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "Rectangle({t}, {l}, {b}, {r})".format(t=self.top, l=self.left, b=self.bottom, r=self.right)


class Polygon:
    def __init__(self, points: List[Point]=None):
        if points is None:
            points = []
        self.points = copy.copy(points)

    def point_in_poly(self, point: Point):
        """Checks if the given point is within the polygon.

        :param point: Point to test
        :return: True if point is within the polygon else False
        """
        n = len(self.points)
        inside = False

        p1x = self.points[0].x
        p1y = self.points[0].y
        for i in range(n+1):
            p = self.points[i % n]
            p2x = p.x
            p2y = p.y
            if point.y > min(p1y, p2y):
                if point.y <= max(p1y, p2y):
                    if point.x <= max(p1x, p2x):
                        if p1y != p2y:
                            xints = (point.y-p1y)*(p2x-p1x)/(p2y-p1y)+p1x
                        if p1x == p2x or point.x <= xints:
                            inside = not inside
            p1x, p1y = p2x, p2y

        return inside

    def random_point(self) -> Point:
        """Returns a random point within this polygon object

        :return: a random point
        """
        # split polygon into triangles using ear clipping
        tris = self.triangulate()

        # calculate areas of the triangles
        areas = [(x, x.area()) for x in tris]
        full_area = sum([x[1] for x in areas])  # calculate full area
        rtri = random.random() * full_area  # get a random value from the area

        # iterate through areas until we found "ours"
        s = areas[0][1]
        i = 1
        while s <= rtri:
            s += areas[i][1]
            i += 1

        # return a random point from this triangle
        return areas[i-1][0].random_point()

    @staticmethod
    def is_convex(a: Point, b: Point, c: Point):
        # only convex if traversing anti-clockwise!
        crossp = (b.x - a.x) * (c.y - a.y) - (b.y - a.y) * (c.x - a.x)
        if crossp >= 0:
            return True
        return False

    @staticmethod
    def in_triangle(a, b, c, p):
        l = [0, 0, 0]
        eps = 0.0000001
        # calculate barycentric coefficients for point p
        # eps is needed as error correction since for very small distances denom->0
        l[0] = ((b.y - c.y) * (p.x - c.x) + (c.x - b.x) * (p.y - c.y)) /\
               (((b.y - c.y) * (a.x - c.x) + (c.x - b.x) * (a.y - c.y)) + eps)
        l[1] = ((c.y - a.y) * (p.x - c.x) + (a.x - c.x) * (p.y - c.y)) /\
               (((b.y - c.y) * (a.x - c.x) + (c.x - b.x) * (a.y - c.y)) + eps)
        l[2] = 1 - l[0] - l[1]
        # check if p lies in triangle (a, b, c)
        for x in l:
            if x >= 1 or x <= 0:
                return False
        return True

    def is_clockwise(self):
        poly_length = len(self.points)
        # initialize sum with last element
        sum_ = (self.points[0].x - self.points[poly_length-1].x) * (self.points[0].y + self.points[poly_length-1].y)
        # iterate over all other elements (0 to n-1)
        for i in range(poly_length-1):
            sum_ += (self.points[i+1].x - self.points[i].x) * (self.points[i+1].y + self.points[i].y)
        return sum_ > 0

    @staticmethod
    def get_ear(poly):
        size = len(poly)
        if size < 3:
            return []
        if size == 3:
            tri = (poly[0], poly[1], poly[2])
            del poly[:]
            return tri
        for i in range(size):
            tritest = False
            p1 = poly[(i-1) % size]
            p2 = poly[i % size]
            p3 = poly[(i+1) % size]
            if Polygon.is_convex(p1, p2, p3):
                for x in poly:
                    if not (x in (p1, p2, p3)) and Polygon.in_triangle(p1, p2, p3, x):
                        tritest = True
                if not tritest:
                    del poly[i % size]
                    return p1, p2, p3
        print('GetEar(): no ear found')
        return []

    def triangulate(self):
        tri = []
        plist = self.points[::-1] if self.is_clockwise() else self.points[:]
        while len(plist) >= 3:
            a = self.get_ear(plist)
            if not a:
                break
            tri.append(Triangle(a))
        return tri

    def outbound_rectangle(self) -> Rectangle:
        top = max([x.x for x in self.points])
        bot = min([x.x for x in self.points])
        right = max([x.y for x in self.points])
        left = min([x.y for x in self.points])
        return Rectangle(top, left, bot, right)

    def __repr__(self):
        return "Polygon([{points}])".format(points=", ".join(map(repr, self.points)))



from typing import Dict
from . import countries
from . import unitgroup
from . import planes
from . import helicopters
from . import ships
from dcs.unit import Vehicle, Plane, Helicopter, Static, Ship
from dcs.point import MovingPoint, StaticPoint
from dcs.country import Country


class Coalition:
    def __init__(self, name, bullseye=None):
        self.name = name
        self.countries = {}  # type: Dict[str, Country]
        self.bullseye = bullseye
        self.nav_points = []  # TODO

    @staticmethod
    def _import_moving_point(mission, group: unitgroup.Group, imp_group) -> unitgroup.Group:
        for imp_point_idx in imp_group["route"]["points"]:
            imp_point = imp_group["route"]["points"][imp_point_idx]
            point = MovingPoint()
            point.load_from_dict(imp_point, mission.translation)
            group.add_point(point)
        return group

    @staticmethod
    def _import_static_point(mission, group: unitgroup.Group, imp_group) -> unitgroup.Group:
        for imp_point_idx in imp_group["route"]["points"]:
            imp_point = imp_group["route"]["points"][imp_point_idx]
            point = StaticPoint()
            point.load_from_dict(imp_point, mission.translation)
            group.add_point(point)
        return group

    def load_from_dict(self, mission, d):
        for country_idx in d["country"]:
            imp_country = d["country"][country_idx]
            _country = countries.get_by_id(imp_country["id"])

            if "vehicle" in imp_country:
                for vgroup_idx in imp_country["vehicle"]["group"]:
                    vgroup = imp_country["vehicle"]["group"][vgroup_idx]
                    vg = unitgroup.VehicleGroup(vgroup["groupId"], mission.translation.get_string(vgroup["name"]),
                                                vgroup["start_time"])
                    vg.load_from_dict(vgroup)
                    mission.current_group_id = max(mission.current_group_id, vg.id)

                    Coalition._import_moving_point(mission, vg, vgroup)

                    # units
                    for imp_unit_idx in vgroup["units"]:
                        imp_unit = vgroup["units"][imp_unit_idx]
                        unit = Vehicle(
                            id=imp_unit["unitId"],
                            name=mission.translation.get_string(imp_unit["name"]),
                            _type=imp_unit["type"])
                        unit.load_from_dict(imp_unit)

                        mission.current_unit_id = max(mission.current_unit_id, unit.id)
                        vg.add_unit(unit)
                    _country.add_vehicle_group(vg)

            if "ship" in imp_country:
                for group_idx in imp_country["ship"]["group"]:
                    imp_group = imp_country["ship"]["group"][group_idx]
                    vg = unitgroup.ShipGroup(imp_group["groupId"], mission.translation.get_string(imp_group["name"]),
                                             imp_group["start_time"])
                    vg.load_from_dict(imp_group)
                    mission.current_group_id = max(mission.current_group_id, vg.id)

                    Coalition._import_moving_point(mission, vg, imp_group)

                    # units
                    for imp_unit_idx in imp_group["units"]:
                        imp_unit = imp_group["units"][imp_unit_idx]
                        unit = Ship(
                            id=imp_unit["unitId"],
                            name=mission.translation.get_string(imp_unit["name"]),
                            _type=ships.ship_map[imp_unit["type"]])
                        unit.load_from_dict(imp_unit)

                        mission.current_unit_id = max(mission.current_unit_id, unit.id)
                        vg.add_unit(unit)
                    _country.add_ship_group(vg)

            if "plane" in imp_country:
                for pgroup_idx in imp_country["plane"]["group"]:
                    pgroup = imp_country["plane"]["group"][pgroup_idx]
                    plane_group = unitgroup.PlaneGroup(pgroup["groupId"],
                                                       mission.translation.get_string(pgroup["name"]),
                                                       pgroup["start_time"])
                    plane_group.load_from_dict(pgroup)
                    mission.current_group_id = max(mission.current_group_id, plane_group.id)

                    Coalition._import_moving_point(mission, plane_group, pgroup)

                    # units
                    for imp_unit_idx in pgroup["units"]:
                        imp_unit = pgroup["units"][imp_unit_idx]
                        plane = Plane(
                            _id=imp_unit["unitId"],
                            name=mission.translation.get_string(imp_unit["name"]),
                            _type=planes.plane_map[imp_unit["type"]],
                            _country=_country)
                        plane.load_from_dict(imp_unit)

                        if plane_group.points[0].airdrome_id is not None and plane.parking is not None:
                            airport = mission.terrain.airport_by_id(plane_group.points[0].airdrome_id)
                            slot = airport.parking_slot(plane.parking)
                            plane.set_parking(slot)

                        mission.current_unit_id = max(mission.current_unit_id, plane.id)
                        plane_group.add_unit(plane)

                    # check runway start
                    if plane_group.points[0].airdrome_id is not None and plane_group.units[0].parking is None:
                        airport = mission.terrain.airport_by_id(plane_group.points[0].airdrome_id)
                        airport.occupy_runway(plane_group)
                    _country.add_plane_group(plane_group)

            if "helicopter" in imp_country:
                for pgroup_idx in imp_country["helicopter"]["group"]:
                    pgroup = imp_country["helicopter"]["group"][pgroup_idx]
                    helicopter_group = unitgroup.HelicopterGroup(
                        pgroup["groupId"],
                        mission.translation.get_string(pgroup["name"]),
                        pgroup["start_time"])
                    helicopter_group.load_from_dict(pgroup)
                    mission.current_group_id = max(mission.current_group_id, helicopter_group.id)

                    Coalition._import_moving_point(mission, helicopter_group, pgroup)

                    # units
                    for imp_unit_idx in pgroup["units"]:
                        imp_unit = pgroup["units"][imp_unit_idx]
                        heli = Helicopter(
                            _id=imp_unit["unitId"],
                            name=mission.translation.get_string(imp_unit["name"]),
                            _type=helicopters.helicopter_map[imp_unit["type"]],
                            _country=_country)
                        heli.load_from_dict(imp_unit)

                        if helicopter_group.points[0].airdrome_id is not None and heli.parking is not None:
                            airport = mission.terrain.airport_by_id(helicopter_group.points[0].airdrome_id)
                            slot = airport.parking_slot(heli.parking)
                            heli.set_parking(slot)

                        mission.current_unit_id = max(mission.current_unit_id, heli.id)
                        helicopter_group.add_unit(heli)

                    # check runway start
                    if helicopter_group.points[0].airdrome_id is not None and helicopter_group.units[0].parking is None:
                        airport = mission.terrain.airport_by_id(helicopter_group.points[0].airdrome_id)
                        airport.occupy_runway(helicopter_group)
                    _country.add_helicopter_group(helicopter_group)

            if "static" in imp_country:
                for sgroup_idx in imp_country["static"]["group"]:
                    sgroup = imp_country["static"]["group"][sgroup_idx]
                    static_group = unitgroup.StaticGroup(sgroup["groupId"],
                                                         mission.translation.get_string(sgroup["name"]))
                    static_group.load_from_dict(sgroup)
                    mission.current_group_id = max(mission.current_group_id, static_group.id)

                    Coalition._import_static_point(mission, static_group, sgroup)

                    # units
                    for imp_unit_idx in sgroup["units"]:
                        imp_unit = sgroup["units"][imp_unit_idx]
                        static = Static(
                            unit_id=imp_unit["unitId"],
                            name=mission.translation.get_string(imp_unit["name"]),
                            _type=imp_unit["type"])
                        static.load_from_dict(imp_unit)

                        mission.current_unit_id = max(mission.current_unit_id, static.id)
                        static_group.add_unit(static)
                    _country.add_static_group(static_group)
            self.add_country(_country)

    def set_bullseye(self, bulls):
        self.bullseye = bulls

    def add_country(self, country):
        self.countries[country.name] = country
        return country

    def remove_country(self, name):
        return self.countries.pop(name)

    def swap_country(self, coalition, name):
        return coalition.add_country(self.remove_country(name))

    def country(self, country_name: str):
        return self.countries.get(country_name, None)

    def find_group(self, group_name, search="exact"):
        for c in self.countries:
            g = self.countries[c].find_group(group_name, search)
            if g:
                return g

        return None

    def dict(self):
        d = {"name": self.name}
        if self.bullseye:
            d["bullseye"] = self.bullseye
        d["country"] = {}
        i = 1
        for country in sorted(self.countries.keys()):
            d["country"][i] = self.country(country).dict()
            i += 1
        d["nav_points"] = {}
        return d



import math
import random
import copy
from enum import Enum
from typing import List, Union
from .unit import Unit, Skill, FlyingUnit, Plane, PlaneType, Helicopter, HelicopterType
from .point import StaticPoint, MovingPoint, PointAction, PointProperties
from .translation import String
from .terrain import Airport, Runway
from . import task
from . import mapping


class Group:
    class Formation(Enum):
        Line = 1
        Star = 2
        Rectangle = 3
        Scattered = 4

    def __init__(self, _id: int, name=None):
        if not isinstance(_id, int):
            raise TypeError("id must be an integer")
        self.id = _id
        self.hidden = False
        self.units = []  # type: List[Unit]
        self.points = []  # type: List[Union[StaticPoint, MovingPoint]
        self.name = name if name else String()

    def load_from_dict(self, d):
        self.hidden = d.get("hidden")

    def add_unit(self, unit: Unit):
        self.units.append(unit)

    def add_point(self, point: StaticPoint):
        self.points.append(point)

    @property
    def x(self):
        if len(self.units) > 0:
            return self.units[0].position.x
        return None


    @property
    def y(self):
        if len(self.units) > 0:
            return self.units[0].position.y
        return None

    @property
    def position(self) -> mapping.Point:
        return self.units[0].position

    def formation_line(self, heading, distance=20):
        pos = self.units[0].position
        for i in range(1, len(self.units)):
            unit = self.units[i]
            unit.position = pos.point_from_heading(heading + 90, i * distance)
            unit.heading = heading

    def formation_star(self, heading, distance=20):
        pos = self.units[0].position
        units_count = len(self.units)
        iterations = math.ceil(units_count / 8)
        u_idx = 1
        for i in range(0, iterations):
            sx = pos.x - (i + 1) * distance
            sy = pos.y - (i + 1) * distance
            for j in range(0, 3):
                dx = distance * (i + 1) * j
                for k in range(0, 3):
                    dy = distance * (i + 1) * k
                    if u_idx >= units_count:
                        break
                    u = self.units[u_idx]
                    u.position.x = sx + dx
                    u.position.y = sy + dy
                    u.heading = heading

                    if not (j == 1 and k == 1):
                        u_idx += 1

    def formation_rectangle(self, heading, distance=20):
        units_count = len(self.units)
        size = math.ceil(math.sqrt(units_count))
        sx = self.units[0].position.x
        sy = self.units[0].position.y
        u_idx = 0
        for i in range(0, size):
            dx = distance * i
            for j in range(0, size):
                dy = distance * j
                if u_idx >= units_count:
                    break
                u = self.units[u_idx]
                u.position.x = sx - dx
                u.position.y = sy + dy
                u.heading = heading
                u_idx += 1

    def formation_scattered(self, heading=0, max_radius=None):
        unit_count = len(self.units)
        max_r = max_radius if max_radius else random.randrange(15, unit_count * 20)

        sx = self.units[0].position.x
        sy = self.units[0].position.y
        start_pos = self.units[0].position

        for i in range(1, unit_count):
            while True:
                pos = start_pos.point_from_heading(random.randrange(0, 360), random.randrange(14, max_r))

                collision = False
                for j in range(0, i):
                    test_unit = self.units[j]
                    unit_rect = mapping.Rectangle.from_point(test_unit.position, 14)

                    if unit_rect.point_in_rect(pos):
                        collision = True

                if not collision:
                    u = self.units[i]
                    u.position = copy.copy(pos)
                    u.heading = heading
                    break
            i += 1

    def formation(self, _type=Formation.Line, heading=0):
        form_map = {
            VehicleGroup.Formation.Line: self.formation_line,
            VehicleGroup.Formation.Star: self.formation_star,
            VehicleGroup.Formation.Rectangle: self.formation_rectangle,
            VehicleGroup.Formation.Scattered: self.formation_scattered
        }

        form_map[_type](heading)

        return True

    def set_skill(self, skill: Skill):
        for u in self.units:
            u.skill = skill

    def dict(self):
        d = {
            "name": self.name.id,
            "groupId": self.id
        }
        if self.hidden is not None:
            d["hidden"] = self.hidden
        if self.units:
            d["x"] = self.units[0].position.x
            d["y"] = self.units[0].position.y
            d["units"] = {}
            i = 1
            for unit in self.units:
                d["units"][i] = unit.dict()
                dunit = d["units"][i]
                if len(self.points) > 1:
                    hdg = self.points[0].position.heading_between_point(self.points[1].position)
                    rhdg = math.radians(hdg)
                    dunit["heading"] = round(rhdg, 13)
                    if "psi" in dunit:
                        dunit["psi"] = round(rhdg * -1, 13)
                i += 1
        if self.points:
            d["route"] = {"points": {}}
            i = 1
            for point in self.points:
                d["route"]["points"][i] = point.dict()
                i += 1
        return d


class MovingGroup(Group):
    def __init__(self, _id, name=None, start_time=0):
        super(MovingGroup, self).__init__(_id, name)
        self.task = ""
        self.tasks = []
        self.start_time = start_time
        self.frequency = 251
        self.task_selected = True
        self.spawn_probability = 1.0

    def load_from_dict(self, d):
        super(MovingGroup, self).load_from_dict(d)
        self.frequency = d.get("frequency")
        self.task = d.get("task")  # ships don't have a task
        self.spawn_probability = d.get("probability", 1.0)
        for t in d.get("tasks", []):
            self.tasks.append(task._create_from_dict(d["task"]["params"]["tasks"][t]))
        self.task_selected = d.get("taskSelected", False)

    def dict(self):
        d = super(MovingGroup, self).dict()
        if self.task:
            d["task"] = self.task
        d["tasks"] = {i+1: self.tasks[i].dict() for i in range(0, len(self.tasks))}
        d["start_time"] = self.start_time
        if self.frequency:
            d["frequency"] = self.frequency
        if self.task_selected:
            d["taskSelected"] = self.task_selected
        if self.spawn_probability != 1.0:
            d["probability"] = self.spawn_probability
        return d


class VehicleGroup(MovingGroup):
    def __init__(self, _id, name=None, start_time=0):
        super(VehicleGroup, self).__init__(_id, name, start_time)
        self.task = "Ground Nothing"
        self.modulation = 0
        self.communication = False
        self.frequency = None
        self.visible = False  # visible before start flag
        self.spans = []

    def load_from_dict(self, d):
        super(VehicleGroup, self).load_from_dict(d)
        self.modulation = d.get("modulation")
        self.communication = d.get("communication", False)
        self.visible = d.get("visible", False)

    def add_span(self, position: mapping.Point):
        self.spans.append({"x": position.x, "y": position.y})

    def add_waypoint(self, position: mapping.Point,
                     move_formation: PointAction=PointAction.OffRoad, speed=32) -> MovingPoint:
        mp = MovingPoint()
        mp.type = "Turning Point"
        mp.action = move_formation
        mp.position = copy.copy(position)
        mp.speed = speed / 3.6
        mp.ETA_locked = False

        self.add_point(mp)
        return mp

    def dict(self):
        d = super(VehicleGroup, self).dict()
        if self.communication:
            d["modulation"] = self.modulation
            d["communication"] = self.communication
            d["frequency"] = self.frequency
        d["visible"] = self.visible
        if self.spans is not None:
            # spans
            d["route"]["spans"] = {}
            i = 1
            for spawn in self.spans:
                d["route"]["spans"][i] = spawn
                i += 1
        return d


class FlyingGroup(MovingGroup):
    def __init__(self, _id, name=None, start_time=0):
        super(FlyingGroup, self).__init__(_id, name, start_time)
        self.modulation = 0
        self.communication = True
        self.uncontrolled = False
        self.radio_set = False
        self.task = "CAS"
        self.units = []  # type: List[FlyingUnit]

    def starts_from_airport(self) -> bool:
        return self.points[0].airdrome_id if self.points else False

    def load_from_dict(self, d):
        super(FlyingGroup, self).load_from_dict(d)
        self.modulation = d.get("modulation")
        self.communication = d.get("communication", False)
        self.uncontrolled = d["uncontrolled"]
        self.radio_set = d.get("radioSet", False)

    def add_waypoint(self, pos: mapping.Point, altitude, speed=600, name: String=None) -> MovingPoint:
        mp = MovingPoint()
        mp.type = "Turning Point"
        mp.action = PointAction.TurningPoint
        mp.name = name if name else String()
        mp.position = mapping.Point(pos.x, pos.y)
        mp.alt = altitude
        mp.speed = speed / 3.6
        mp.ETA_locked = False
        mp.properties = PointProperties()

        self.add_point(mp)
        return mp

    def add_runway_waypoint(self, airport: Airport, runway: Runway=None,
                            distance=random.randrange(6000, 8000, 100)) -> MovingPoint:
        """Adds a waypoint parallel to the given runway heading, for start or approach.

        :param airport: start airport object
        :param runway: runway for heading direction, if None first(default) airport runway will be used.
        :param distance: distance of the waypoint from the airport
        :return: MovePoint object describing the waypoint
        """
        runway = runway if runway else airport.runways[0]

        mp = MovingPoint()
        mp.type = "Turning Point"
        mp.action = PointAction.TurningPoint
        mp.alt_type = "RADIO"
        mp.position = airport.position.point_from_heading(runway.heading, distance)
        mp.alt = 300
        mp.speed = 200 / 3.6
        mp.ETA_locked = False
        mp.properties = PointProperties()

        self.add_point(mp)
        return mp

    def land_at(self, airport: Airport) -> MovingPoint:
        mp = MovingPoint()
        mp.type = "Land"
        mp.action = PointAction.Landing
        mp.position = copy.copy(airport.position)
        mp.airdrome_id = airport.id
        mp.alt = 0
        mp.speed = 0
        mp.ETA_locked = False
        mp.properties = PointProperties()

        self.add_point(mp)
        return mp

    def load_task_default_loadout(self, task):
        task_payload = self.units[0].unit_type.loadout(task)
        if task_payload:
            for p in self.units:
                for x in task_payload:
                    p.load_pylon(x)

    def load_loadout(self, name):
        payload = self.units[0].unit_type.loadout_by_name(name)
        if payload:
            for p in self.units:
                for x in payload:
                    p.load_pylon(x)

    def load_pylon(self, weapon, pylon=None):
        for u in self.units:
            u.load_pylon(weapon, pylon)

        return True

    def load_loadout_from_file(self, filename):
        for u in self.units:
            u.load_loadout(filename)

        return True

    def set_client(self):
        for u in self.units:
            u.set_client()

    def reset_loadout(self):
        for u in self.units:
            u.reset_loadout()

    def set_frequency(self, frequency):
        self.frequency = frequency
        self.radio_set = True

    def dict(self):
        # if a player/client is in the group
        # make sure his 1. preset channel is at frequency
        for u in self.units:
            if u.skill in [Skill.Client, Skill.Player]:
                u.set_default_preset_channel(self.frequency)

        d = super(FlyingGroup, self).dict()
        d["modulation"] = self.modulation
        d["communication"] = self.communication
        d["uncontrolled"] = self.uncontrolled
        d["radioSet"] = self.radio_set

        return d


class PlaneGroup(FlyingGroup):
    def __init__(self, _id, name=None, start_time=0):
        super(PlaneGroup, self).__init__(_id, name, start_time)

    def add_unit(self, unit: Plane):
        if not issubclass(unit.unit_type, PlaneType):
            print(unit.unit_type)
            raise TypeError("unit.unit_type is not a plane")
        super(PlaneGroup, self).add_unit(unit)


class HelicopterGroup(FlyingGroup):
    def __init__(self, _id, name=None, start_time=0):
        super(HelicopterGroup, self).__init__(_id, name, start_time)
        self.frequency = 127.5

    def add_unit(self, unit: Helicopter):
        if not issubclass(unit.unit_type, HelicopterType):
            raise TypeError("unit.unit_type is not a helicopter")
        super(HelicopterGroup, self).add_unit(unit)


class ShipGroup(MovingGroup):
    def __init__(self, _id, name=None, start_time=0):
        super(ShipGroup, self).__init__(_id, name, start_time)
        self.task = None
        self.frequency = None
        self.spans = None
        self.visible = False  # visible before start flag

    def add_waypoint(self, position: mapping.Point, speed=20) -> MovingPoint:
        mp = MovingPoint()
        mp.type = "Turning Point"
        mp.action = PointAction.TurningPoint
        mp.position = copy.copy(position)
        mp.speed = speed / 3.6
        mp.ETA_locked = False

        self.add_point(mp)
        return mp

    def dict(self):
        d = super(ShipGroup, self).dict()
        d["visible"] = self.visible
        return d


class StaticGroup(Group):
    def __init__(self, _id, name=None):
        super(StaticGroup, self).__init__(_id, name)
        self.dead = False
        self.heading = 0

    def load_from_dict(self, d):
        super(StaticGroup, self).load_from_dict(d)
        self.heading = math.degrees(d["heading"])
        self.dead = d["dead"]

    def dict(self):
        d = super(StaticGroup, self).dict()
        d["dead"] = self.dead
        d["heading"] = round(math.radians(self.heading), 13)
        return d



class GroundControl:
    def __init__(self):
        self.pilot_can_control_vehicles = False
        self.red_game_masters = 0
        self.red_tactical_commander = 0
        self.red_jtac = 0
        self.red_observer = 0

        self.blue_game_masters = 0
        self.blue_tactical_commander = 0
        self.blue_jtac = 0
        self.blue_observer = 0

    def load_from_dict(self, d):
        if d is None:
            return
        self.pilot_can_control_vehicles = d["isPilotControlVehicles"]

        self.red_game_masters = int(d["roles"]["instructor"]["red"])
        self.red_tactical_commander = int(d["roles"]["artillery_commander"]["red"])
        self.red_jtac = int(d["roles"]["forward_observer"]["red"])
        self.red_observer = int(d["roles"]["observer"]["red"])

        self.blue_game_masters = int(d["roles"]["instructor"]["blue"])
        self.blue_tactical_commander = int(d["roles"]["artillery_commander"]["blue"])
        self.blue_jtac = int(d["roles"]["forward_observer"]["blue"])
        self.blue_observer = int(d["roles"]["observer"]["blue"])

    def dict(self):
        return {
            "isPilotControlVehicles": self.pilot_can_control_vehicles,
            "roles": {
                "instructor": {"red": self.red_game_masters, "blue": self.blue_game_masters},
                "artillery_commander": {"red": self.red_tactical_commander, "blue": self.blue_tactical_commander},
                "forward_observer": {"red": self.red_jtac, "blue": self.blue_jtac},
                "observer": {"red": self.red_observer, "blue": self.blue_observer},
            }
        }



from .translation import String
from . import task
from . import mapping
from typing import List
from enum import Enum


class PointAction(Enum):
    None_ = ""
    TurningPoint = "Turning Point"
    FlyOverPoint = "Fly Over Point"
    FromParkingArea = "From Parking Area"
    FromParkingAreaHot = "From Parking Area Hot"
    FromRunway = "From Runway"
    Landing = "Landing"
    OnRoad = "On Road"
    OffRoad = "Off Road"
    LineAbreast = "Rank"
    Cone = "Cone"
    Vee = "Vee"
    Diamond = "Diamond"
    EchelonLeft = "EchelonL"
    EchelonRight = "EchelonR"
    Custom = "Custom"


class StaticPoint:
    def __init__(self):
        self.alt = 0
        self.type = ""
        self.name = String()
        self.position = mapping.Point(0, 0)
        self.speed = 0
        self.formation_template = ""
        self.action = PointAction.None_  # type: PointAction

    def load_from_dict(self, d, translation):
        self.alt = d["alt"]
        self.type = d["type"]
        self.position.x = d["x"]
        self.position.y = d["y"]
        self.action = PointAction(d["action"])
        self.formation_template = d["formation_template"]
        self.speed = d["speed"]
        self.name = translation.get_string(d["name"])

    def dict(self):
        return {
            "alt": self.alt,
            "type": self.type,
            "name": self.name.id,
            "x": self.position.x,
            "y": self.position.y,
            "speed": round(self.speed, 13),
            "formation_template": self.formation_template,
            "action": self.action.value
        }


class VNav(Enum):
    V2D = 0
    V3D = 1
    VNone = 2


class Scale(Enum):
    Enroute = 0
    Terminal = 1
    Approach = 2
    HighAcc = 3
    None_ = 4


class Steer(Enum):
    ToFrom = 0
    Direct = 1
    ToTo = 2
    None_ = 3


class PointProperties:
    def __init__(self, vnav: VNav=VNav.V2D, scale: Scale=Scale.Enroute, steer: Steer=Steer.ToTo, angle=None):
        self.vnav = vnav
        self.scale = scale
        self.steer = steer
        if angle is not None:
            self.angle = 1
            self.vangle = angle
        else:
            self.angle = 0
            self.vangle = 0

    def load_from_dict(self, d):
        self.vnav = VNav(d["vnav"])
        self.scale = Scale(d["scale"])
        self.steer = Steer(d["steer"])
        self.angle = d["angle"]
        self.vangle = d["vangle"]

    def dict(self):
        return {
            "vnav": self.vnav.value,
            "scale": self.scale.value,
            "angle": self.angle,
            "vangle": self.vangle,
            "steer": self.steer.value
        }


class MovingPoint(StaticPoint):
    def __init__(self):
        super(MovingPoint, self).__init__()
        self.type = "Turning Point"
        self.action = PointAction.TurningPoint  # type: PointAction
        self.alt_type = "BARO"
        self.ETA = 0
        self.ETA_locked = True
        self.speed_locked = True
        self.tasks = []  # type: List[task.Task]
        self.properties = None  # type: PointProperties
        self.airdrome_id = None
        self.helipad_id = None

    def load_from_dict(self, d, translation):
        super(MovingPoint, self).load_from_dict(d, translation)
        self.alt_type = d.get("alt_type", None)
        self.ETA_locked = d["ETA_locked"]
        self.ETA = d["ETA"]
        self.speed_locked = d["speed_locked"]
        for t in d["task"]["params"]["tasks"]:
            self.tasks.append(task._create_from_dict(d["task"]["params"]["tasks"][t]))
        self.airdrome_id = d.get("airdromeId", None)
        self.helipad_id = d.get("helipadId", None)
        if d.get("properties"):
            self.properties = PointProperties()
            self.properties.load_from_dict(d.get("properties"))
        else:
            self.properties = None

    def find_task(self, task_type):
        """Searches tasks in this point for the given task class

        :param task_type: task class to search, :py:mod:`dcs.task`
        :return: task instance if found, else None
        """
        for t in self.tasks:
            if isinstance(t, task_type):
                return t
        return None

    def dict(self):
        d = super(MovingPoint, self).dict()
        d["alt_type"] = self.alt_type
        d["ETA"] = self.ETA
        d["ETA_locked"] = self.ETA_locked
        d["speed_locked"] = self.speed_locked
        tasks = {}
        for i in range(0, len(self.tasks)):
            self.tasks[i].number = i+1
            tasks[i+1] = self.tasks[i].dict()
        d["task"] = {
            "id": "ComboTask",
            "params": {
                "tasks": {i+1: self.tasks[i].dict() for i in range(0, len(self.tasks))}
            }
        }
        if self.airdrome_id is not None:
            d["airdromeId"] = self.airdrome_id
        if self.helipad_id is not None:
                d["helipadId"] = self.helipad_id
        if self.properties is not None:
            d["properties"] = self.properties.dict()
        return d



from typing import Dict


class String:
    def __init__(self, _id='', translation=None):
        self.translation = translation
        self.id = _id

    def __del__(self):
        if self.translation:
            self.translation.delete_string(self.id)

    def set(self, text, lang='DEFAULT'):
        self.translation.set_string(self.id, text, lang)
        return str(self)

    def str(self, lang='DEFAULT'):
        return self.translation.strings[lang][self.id]

    def __str__(self):
        return self.str('DEFAULT')

    def __repr__(self):
        return self.id + ":" + str(self)


class Translation:
    def __init__(self, _mission):
        self.strings = {}  # type: Dict[str,Dict[str,str]]
        self.mission = _mission

    def set_string(self, _id, string, lang='DEFAULT'):
        if lang not in self.strings:
            self.strings[lang] = {}
        self.strings[lang][_id] = string
        return _id

    def get_string(self, _id):
        return String(_id, self)

    def create_string(self, s, lang='DEFAULT'):
        _id = 'DictKey_Translation_{dict_id}'.format(dict_id=self.mission.next_dict_id())
        self.set_string(_id, s, lang)
        return String(_id, self)

    def delete_string(self, _id):
        for lang in self.strings:
            if _id in self.strings[lang]:
                del self.strings[lang][_id]

    def languages(self) -> [str]:
        return self.strings.keys()

    def dict(self, lang='DEFAULT'):
        if lang in self.strings:
            return {x: self.strings[lang][x] for x in self.strings[lang]}
        return {}

    def __str__(self):
        return str(self.strings)

    def __repr__(self):
        return repr(self.strings)



def dumps(value, varname=None, indent=None):
    nl = "\n" if indent else ""
    s = varname + '=' + nl if varname else ''
    indentcount = 0 if indent is None else indent

    if isinstance(value, dict):
        e = []
        for key in sorted(value.keys(), key=str):
            child = value[key]
            KNL = "\n" if indent and isinstance(child, (dict, list)) else ""
            selem = '\t' * indentcount
            skey = key if isinstance(key, int) else '"{key}"'.format(key=key)
            selem += '[{key}]={nl}'.format(key=skey, nl=KNL)
            selem += dumps(value[key], indent=indent + 1 if indent else None)
            e.append(selem)
        s += '\t' * (indent-1) if nl else ""
        s += "{"
        if e:
            s += nl + ",{nl}".format(nl=nl).join(e)
        s += nl + '\t' * (indentcount-1) + "}"
    elif isinstance(value, list):
        e = []
        i = 1
        for v in value:
            selem = '\t' * indentcount + "[{i}]=".format(i=i)
            selem += dumps(v, indent=indent + 1 if indent else None)
            e.append(selem)
            i += 1
        s += '\t' * (indent-1) if nl else ""
        s += "{"
        if e:
            s += nl + ",{nl}".format(nl=nl).join(e)
        s += nl + '\t' * (indentcount-1) + "}"
    elif isinstance(value, str):
        v = value.replace('\\', '\\\\')
        v = v.replace('"', '\\"')
        v = v.replace('\n', '\\\n')
        s += '"{val}"'.format(val=v)
    elif isinstance(value, bool):
        s += "true" if value else "false"
    else:
        s += str(value)

    return s
